"""OPP serial communicator."""
import asyncio

from mpf.platforms.opp.opp_rs232_intf import OppRs232Intf

from mpf.platforms.base_serial_communicator import BaseSerialCommunicator

MYPY = False
if MYPY:    # pragma: no cover
    from mpf.platforms.opp.opp import OppHardwarePlatform   # pylint: disable-msg=cyclic-import,unused-import

# Minimum firmware versions needed for this module
MIN_FW = 0x00000100
BAD_FW_VERSION = 0x01020304


class OPPSerialCommunicator(BaseSerialCommunicator):

    """Manages a Serial connection to the first processor in a OPP serial chain."""

    __slots__ = ["part_msg", "chain_serial", "_lost_synch"]

    # pylint: disable=too-many-arguments
    def __init__(self, platform: "OppHardwarePlatform", port, baud, overwrite_serial) -> None:
        """Initialise Serial Connection to OPP Hardware."""
        self.part_msg = b""
        self.chain_serial = overwrite_serial    # type: str
        self._lost_synch = False

        super().__init__(platform, port, baud)
        self.platform = platform    # hint the right type

    async def _read_id(self):
        msg = bytearray([0x20, 0x00, 0x00, 0x00, 0x00, 0x00])
        msg.extend(OppRs232Intf.calc_crc8_whole_msg(msg))
        msg.extend(OppRs232Intf.EOM_CMD)
        self.send(bytes(msg))

        resp = await self.read(8)
        if resp[7] != ord(OppRs232Intf.EOM_CMD):
            raise AssertionError("Failed to read ID from {}. Missing EOM.".format(self.port))

        if ord(OppRs232Intf.calc_crc8_whole_msg(resp[0:6])) != resp[6]:
            raise AssertionError("Failed to read ID from {}. Wrong CRC.".format(self.port))

        if resp[0] != 0x20:
            raise AssertionError("Failed to read ID from {}. Wrong Board ID.".format(self.port))

        if resp[1] != 0:
            raise AssertionError("Failed to read ID from {}. Wrong CMD.".format(self.port))

        return (resp[2] << 24) + (resp[3] << 16) + (resp[4] << 8) + resp[5]

    async def _identify_connection(self):
        """Identify which processor this serial connection is talking to."""
        # keep looping and wait for an ID response
        count = 0
        # read and discard all messages in buffer
        self.send(OppRs232Intf.EOM_CMD)
        await asyncio.sleep(.01)
        await self.read(1000)
        while True:
            if (count % 10) == 0:
                self.log.debug("Sending EOM command to port '%s'",
                               self.port)
            count += 1
            self.send(OppRs232Intf.EOM_CMD)
            await asyncio.sleep(.01)
            resp = await self.read(30)
            if resp.startswith(OppRs232Intf.EOM_CMD):
                break
            if count == 100:
                raise AssertionError('No response from OPP hardware: {}'.format(self.port))

        self.log.debug("Got ID response: %s", "".join(" 0x%02x" % b for b in resp))
        if self.chain_serial is None:
            # get ID from hardware if it is not overwritten
            self.chain_serial = str(await self._read_id())

        if self.chain_serial in self.platform.opp_connection:
            raise AssertionError("Duplicate chain serial {} on ports: {} and {}. Each OPP board has to have a "
                                 "unique ID. You can overwrite this using the chains "
                                 "setting.".format(self.chain_serial, self.port,
                                                   self.platform.opp_connection[self.chain_serial]))

        # Send inventory command to figure out number of cards
        msg = bytearray()
        msg.extend(OppRs232Intf.INV_CMD)
        msg.extend(OppRs232Intf.EOM_CMD)
        cmd = bytes(msg)

        self.log.debug("Sending inventory command: %s", "".join(" 0x%02x" % b for b in cmd))
        self.send(cmd)

        resp = await self.readuntil(b'\xff')

        # resp will contain the inventory response.
        self.platform.process_received_message(self.chain_serial, resp)

        # Now send get gen2 configuration message to find populated wing boards
        self.send_get_gen2_cfg_cmd()
        resp = await self.readuntil(b'\xff', 6)

        # resp will contain the gen2 cfg responses.  That will end up creating all the
        # correct objects.
        self.platform.process_received_message(self.chain_serial, resp)

        # get the version of the firmware
        self.send_vers_cmd()
        resp = await self.readuntil(b'\xff', 6)
        self.platform.process_received_message(self.chain_serial, resp)

        # see if version of firmware is new enough
        if self.platform.min_version[self.chain_serial] < MIN_FW:
            raise AssertionError("Firmware version mismatch. MPF requires"
                                 " the OPP Gen2 processor to be firmware {}, but yours is {}".
                                 format(self._create_vers_str(MIN_FW),
                                        self._create_vers_str(self.platform.min_version[self.chain_serial])))

        # get initial value for inputs
        self.log.debug("Getting initial inputs states for %s", self.chain_serial)
        self.send(self.platform.read_input_msg[self.chain_serial])
        cards = len([x for x in self.platform.opp_inputs if x.chain_serial == self.chain_serial])
        while True:
            resp = await self.readuntil(b'\xff')
            cards -= self._parse_msg(resp)
            if cards <= 0:
                break
            self.log.debug("Waiting for another %s cards", cards)

        self.log.info("Init of OPP board %s done", self.chain_serial)
        self.platform.register_processor_connection(self.chain_serial, self)

    def send_get_gen2_cfg_cmd(self):
        """Send get gen2 configuration message to find populated wing boards."""
        whole_msg = bytearray()
        for card_addr in self.platform.gen2_addr_arr[self.chain_serial]:
            msg = bytearray()
            msg.append(card_addr)
            msg.extend(OppRs232Intf.GET_GEN2_CFG)
            msg.append(0)
            msg.append(0)
            msg.append(0)
            msg.append(0)
            msg.extend(OppRs232Intf.calc_crc8_whole_msg(msg))
            whole_msg.extend(msg)

        whole_msg.extend(OppRs232Intf.EOM_CMD)
        cmd = bytes(whole_msg)
        self.log.debug("Sending get Gen2 Cfg command: %s", "".join(" 0x%02x" % b for b in cmd))
        self.send(cmd)

    def send_vers_cmd(self):
        """Send get firmware version message."""
        whole_msg = bytearray()
        for card_addr in self.platform.gen2_addr_arr[self.chain_serial]:
            msg = bytearray()
            msg.append(card_addr)
            msg.extend(OppRs232Intf.GET_VERS_CMD)
            msg.append(0)
            msg.append(0)
            msg.append(0)
            msg.append(0)
            msg.extend(OppRs232Intf.calc_crc8_whole_msg(msg))
            whole_msg.extend(msg)

        whole_msg.extend(OppRs232Intf.EOM_CMD)
        cmd = bytes(whole_msg)
        self.log.debug("Sending get version command: %s", "".join(" 0x%02x" % b for b in cmd))
        self.send(cmd)

    @classmethod
    def _create_vers_str(cls, version_int):     # pragma: no cover
        return ("%02d.%02d.%02d.%02d" % (((version_int >> 24) & 0xff),
                                         ((version_int >> 16) & 0xff), ((version_int >> 8) & 0xff),
                                         (version_int & 0xff)))

    def lost_synch(self):
        """Mark connection as desynchronised."""
        self._lost_synch = True

    def _parse_msg(self, msg):
        self.part_msg += msg
        strlen = len(self.part_msg)
        message_found = 0
        # Split into individual responses
        while strlen > 2:
            if self._lost_synch:
                while strlen > 0:
                    # wait for next gen2 card message
                    if (self.part_msg[0] & 0xe0) == 0x20:
                        self._lost_synch = False
                        break
                    self.part_msg = self.part_msg[1:]
                    strlen -= 1
            # Check if this is a gen2 card address
            elif (self.part_msg[0] & 0xe0) == 0x20:
                # Check if read input
                if self.part_msg[1] == ord(OppRs232Intf.READ_GEN2_INP_CMD):
                    if strlen >= 7:
                        self.platform.process_received_message(self.chain_serial, self.part_msg[:7])
                        message_found += 1
                        self.part_msg = self.part_msg[7:]
                        strlen -= 7
                    else:
                        # message not complete yet
                        break
                # Check if read matrix input
                elif self.part_msg[1] == ord(OppRs232Intf.READ_MATRIX_INP):
                    if strlen >= 11:
                        self.platform.process_received_message(self.chain_serial, self.part_msg[:11])
                        message_found += 1
                        self.part_msg = self.part_msg[11:]
                        strlen -= 11
                    else:
                        # message not complete yet
                        break
                else:
                    # Lost synch
                    self.part_msg = self.part_msg[2:]
                    strlen -= 2
                    self._lost_synch = True

            elif self.part_msg[0] == ord(OppRs232Intf.EOM_CMD):
                self.part_msg = self.part_msg[1:]
                strlen -= 1
            else:
                # Lost synch
                self.part_msg = self.part_msg[1:]
                strlen -= 1
                self._lost_synch = True

        return message_found



"""Support for physical segment displays."""
import abc
from typing import Any
from enum import Enum


class FlashingType(Enum):

    """Determine how a segment display should flash."""

    NO_FLASH = False
    FLASH_ALL = True
    FLASH_MATCH = "match"


class SegmentDisplayPlatformInterface(metaclass=abc.ABCMeta):

    """Interface for a segment display in hardware platforms."""

    __slots__ = ["number"]

    def __init__(self, number: Any) -> None:
        """Remember the number."""
        self.number = number

    @abc.abstractmethod
    def set_text(self, text: str, flashing: FlashingType) -> None:
        """Set a text to the display.

        This text will be right aligned in case the text is shorter than the display.
        If it is too long it will be cropped on the left.
        """
        raise NotImplementedError


class SegmentDisplaySoftwareFlashPlatformInterface(SegmentDisplayPlatformInterface):

    """SegmentDisplayPlatformInterface with software emulation for flashing."""

    __slots__ = ["_flash_on", "_flashing", "_text"]

    def __init__(self, number: Any) -> None:
        """Remember the number."""
        super().__init__(number)
        self._flash_on = True
        self._flashing = FlashingType.NO_FLASH      # type: FlashingType
        self._text = ""

    def set_software_flash(self, state):
        """Set software flashing state."""
        self._flash_on = state

        if self._flashing == FlashingType.NO_FLASH:
            return

        # do not flash empty text
        if not self._text:
            return

        if state:
            self._set_text(self._text)
        else:
            if self._flashing == FlashingType.FLASH_MATCH:
                # blank the last two chars
                self._set_text(self._text[0:-2] + "  ")
            else:
                self._set_text("")

    def set_text(self, text: str, flashing: FlashingType) -> None:
        """Set a text to the display."""
        self._text = text
        self._flashing = flashing
        if flashing == FlashingType.NO_FLASH:
            self._flash_on = True
        if flashing == FlashingType.NO_FLASH or self._flash_on or not text:
            self._set_text(text)

    @abc.abstractmethod
    def _set_text(self, text: str) -> None:
        """Set a text to the display."""
        raise NotImplementedError



"""Test high score mode."""
from collections import OrderedDict

from unittest.mock import MagicMock, call
from mpf.tests.MpfBcpTestCase import MpfBcpTestCase
from mpf.tests.MpfTestCase import test_config_directory


class TestHighScoreMode(MpfBcpTestCase):

    def get_config_file(self):
        return 'high_score.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/high_score/'

    def start_game(self, num_players=1):
        self._bcp_client.send = MagicMock()

        self.machine.playfield.add_ball = MagicMock()
        self.machine.events.post('game_start')
        self.advance_time_and_run()
        self.machine.game.balls_in_play = 1
        self.assertIsNotNone(self.machine.game)

        while self.machine.game.num_players < num_players:
            self.machine.game.request_player_add()
            self.advance_time_and_run()

    def test_high_score_without_game(self):
        # try to start high score mode without a running game
        self.assertIn("high_score", self.machine.modes)
        self.assertModeNotRunning("high_score")
        self.machine.modes["high_score"].start()
        self.advance_time_and_run()
        # mode should not start (neither crash)
        self.assertModeNotRunning("high_score")

    def test_default_high_scores(self):
        new_score_data = OrderedDict()
        new_score_data['score'] = [('BRI', 4242),
                                   ('GHK', 2323),
                                   ('JK', 1337),
                                   ('QC', 42),
                                   ('MPF', 23)]
        new_score_data['loops'] = [('JK', 42)]

        self.assertEqual(new_score_data, self.machine.modes["high_score"].high_scores)

    def test_no_high_scores(self):
        self.machine.modes["high_score"].high_scores = OrderedDict(
            score=[('BRI', 7050550),
                   ('GHK', 93060),
                   ('JK', 87890),
                   ('QC', 87890),
                   ('MPF', 10000)])

        self.start_game(4)
        self.machine.game.player_list[0].score = 1000
        self.machine.game.player_list[1].score = 1000
        self.machine.game.player_list[2].score = 1000
        self.machine.game.player_list[3].score = 1000
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.advance_time_and_run()
        self.assertFalse(self.machine.modes["high_score"].active)

        new_score_data = OrderedDict()
        new_score_data['score'] = [('BRI', 7050550),
                                   ('GHK', 93060),
                                   ('JK', 87890),
                                   ('QC', 87890),
                                   ('MPF', 10000)]
        new_score_data['loops'] = []
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    def test_tilt_during_high_score(self):
        self.mock_event("high_score_enter_initials")
        self.machine.modes["high_score"].high_scores = OrderedDict(
            score=[('BRI', 7050550),
                   ('GHK', 93060),
                   ('JK', 87890),
                   ('QC', 87890),
                   ('MPF', 1000)])

        self.start_game()
        self.advance_time_and_run()
        self.machine.game.player_list[0].score = 8000000
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)

        self.assertEqual(1, self._events['high_score_enter_initials'])

        # tilt the machine
        self.hit_and_release_switch("s_tilt")
        self.advance_time_and_run()
        self.hit_and_release_switch("s_tilt")
        self.advance_time_and_run()
        self.hit_and_release_switch("s_tilt")
        self.advance_time_and_run()

        # high score should not end
        self.assertTrue(self.machine.modes["high_score"].active)

    def test_reset(self):
        """Test high score reset."""
        self.assertMachineVarEqual('GRAND CHAMPION', "score1_label")
        self.assertMachineVarEqual(1000000, "score1_value")
        self.assertMachineVarEqual('ABC', "score1_name")
        self.assertMachineVarEqual('HIGH SCORE 1', "score2_label")
        self.assertMachineVarEqual(900000, "score2_value")
        self.assertMachineVarEqual('ABC', "score2_name")
        self.assertMachineVarEqual('HIGH SCORE 2', "score3_label")
        self.assertMachineVarEqual(800000, "score3_value")
        self.assertMachineVarEqual('ABC', "score3_name")
        self.assertMachineVarEqual('HIGH SCORE 3', "score4_label")
        self.assertMachineVarEqual(700000, "score4_value")
        self.assertMachineVarEqual('ABC', "score4_name")
        self.assertMachineVarEqual('HIGH SCORE 4', "score5_label")
        self.assertMachineVarEqual(600000, "score5_value")
        self.assertMachineVarEqual('ABC', "score5_name")
        self.assertMachineVarEqual('LOOP CHAMP', "loops1_label")
        self.assertMachineVarEqual(500000, "loops1_value")
        self.assertMachineVarEqual('MPF', "loops1_name")

        self.assertModeNotRunning("high_score")

        self.post_event("high_scores_reset")
        self.advance_time_and_run(.1)

        new_score_data = OrderedDict()
        new_score_data['score'] = [('BRI', 4242),
                                   ('GHK', 2323),
                                   ('JK', 1337),
                                   ('QC', 42),
                                   ('MPF', 23)]
        new_score_data['loops'] = [('JK', 42)]

        self.assertEqual(new_score_data, self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

        self.assertMachineVarEqual('GRAND CHAMPION', "score1_label")
        self.assertMachineVarEqual(4242, "score1_value")
        self.assertMachineVarEqual('BRI', "score1_name")
        self.assertMachineVarEqual('HIGH SCORE 1', "score2_label")
        self.assertMachineVarEqual(2323, "score2_value")
        self.assertMachineVarEqual('GHK', "score2_name")
        self.assertMachineVarEqual('HIGH SCORE 2', "score3_label")
        self.assertMachineVarEqual(1337, "score3_value")
        self.assertMachineVarEqual('JK', "score3_name")
        self.assertMachineVarEqual('HIGH SCORE 3', "score4_label")
        self.assertMachineVarEqual(42, "score4_value")
        self.assertMachineVarEqual('QC', "score4_name")
        self.assertMachineVarEqual('HIGH SCORE 4', "score5_label")
        self.assertMachineVarEqual(23, "score5_value")
        self.assertMachineVarEqual('MPF', "score5_name")
        self.assertMachineVarEqual('LOOP CHAMP', "loops1_label")
        self.assertMachineVarEqual(42, "loops1_value")
        self.assertMachineVarEqual('JK', "loops1_name")

        self.mock_event("high_score_enter_initials")
        self.start_game()
        self.advance_time_and_run()
        self.machine.game.player_list[0].score = 8000000
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)
        self.assertEqual(1, self._events['high_score_enter_initials'])
        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete', text='NEW')))

        self.advance_time_and_run(5)

        self.assertMachineVarEqual('GRAND CHAMPION', "score1_label")
        self.assertMachineVarEqual(8000000, "score1_value")
        self.assertMachineVarEqual('NEW', "score1_name")
        self.assertMachineVarEqual('HIGH SCORE 1', "score2_label")
        self.assertMachineVarEqual(4242, "score2_value")
        self.assertMachineVarEqual('BRI', "score2_name")
        self.assertMachineVarEqual('HIGH SCORE 2', "score3_label")
        self.assertMachineVarEqual(2323, "score3_value")
        self.assertMachineVarEqual('GHK', "score3_name")
        self.assertMachineVarEqual('HIGH SCORE 3', "score4_label")
        self.assertMachineVarEqual(1337, "score4_value")
        self.assertMachineVarEqual('JK', "score4_name")
        self.assertMachineVarEqual('HIGH SCORE 4', "score5_label")
        self.assertMachineVarEqual(42, "score5_value")
        self.assertMachineVarEqual('QC', "score5_name")
        self.assertMachineVarEqual('LOOP CHAMP', "loops1_label")
        self.assertMachineVarEqual(42, "loops1_value")
        self.assertMachineVarEqual('JK', "loops1_name")

        # test reset after a game
        self.post_event("high_scores_reset")
        self.advance_time_and_run(.1)

        self.assertMachineVarEqual('GRAND CHAMPION', "score1_label")
        self.assertMachineVarEqual(4242, "score1_value")
        self.assertMachineVarEqual('BRI', "score1_name")
        self.assertMachineVarEqual('HIGH SCORE 1', "score2_label")
        self.assertMachineVarEqual(2323, "score2_value")
        self.assertMachineVarEqual('GHK', "score2_name")
        self.assertMachineVarEqual('HIGH SCORE 2', "score3_label")
        self.assertMachineVarEqual(1337, "score3_value")
        self.assertMachineVarEqual('JK', "score3_name")
        self.assertMachineVarEqual('HIGH SCORE 3', "score4_label")
        self.assertMachineVarEqual(42, "score4_value")
        self.assertMachineVarEqual('QC', "score4_name")
        self.assertMachineVarEqual('HIGH SCORE 4', "score5_label")
        self.assertMachineVarEqual(23, "score5_value")
        self.assertMachineVarEqual('MPF', "score5_name")
        self.assertMachineVarEqual('LOOP CHAMP', "loops1_label")
        self.assertMachineVarEqual(42, "loops1_value")
        self.assertMachineVarEqual('JK', "loops1_name")

        self.assertEqual(new_score_data, self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    def test_add_player_during_high_score(self):
        self.mock_event("high_score_enter_initials")
        self.machine.modes["high_score"].high_scores = OrderedDict(
            score=[('BRI', 7050550),
                   ('GHK', 93060),
                   ('JK', 87890),
                   ('QC', 87890),
                   ('MPF', 1000)])

        self.start_game()
        self.advance_time_and_run()
        self.machine.game.player_list[0].score = 8000000
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)

        self.assertEqual(1, self._events['high_score_enter_initials'])
        self.machine.game.request_player_add()
        self.advance_time_and_run()

        # player should not be added
        self.assertEqual(len(self.machine.game.player_list), 1)

    def test_1_high_score(self):
        self.mock_event("high_score_enter_initials")
        self.machine.modes["high_score"].high_scores = OrderedDict(
            score=[('BRI', 7050550),
                   ('GHK', 93060),
                   ('JK', 87890),
                   ('QC', 87890),
                   ('MPF', 1000)])

        self.start_game(4)
        self.advance_time_and_run()
        self.machine.game.player_list[0].score = 8000000
        self.machine.game.player_list[1].score = 1000
        self.machine.game.player_list[2].score = 1000
        self.machine.game.player_list[3].score = 1000
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)

        self.assertEqual(1, self._events['high_score_enter_initials'])

        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete', text='NEW')))

        # award slide display time is 4 secs
        self.advance_time_and_run(2)

        # make sure the high score mode is still running
        self.assertTrue(self.machine.modes["high_score"].active)

        # another 2 secs and it should be done
        self.advance_time_and_run(3)
        self.assertFalse(self.machine.modes["high_score"].active)

        new_score_data = OrderedDict()
        new_score_data['score'] = [('NEW', 8000000),
                                   ('BRI', 7050550),
                                   ('GHK', 93060),
                                   ('JK', 87890),
                                   ('QC', 87890)]
        new_score_data['loops'] = []

        self.assertEqual(new_score_data,
                         self.machine.modes["high_score"].high_scores)

        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    def test_2_high_scores(self):
        self.mock_event("high_score_enter_initials")
        self.machine.modes["high_score"].high_scores = OrderedDict(
            score=[('BRI', 7050550),
                   ('GHK', 93060),
                   ('JK', 87890),
                   ('QC', 87890),
                   ('MPF', 1000)])

        self.start_game(4)
        self.machine.game.player_list[0].score = 8000000
        self.machine.game.player_list[1].score = 10000000
        self.machine.game.player_list[2].score = 1000
        self.machine.game.player_list[3].score = 1000
        self._bcp_client.send.reset_mock()
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)

        # GC

        self.assertEqual(1, self._events['high_score_enter_initials'])
        self._bcp_client.send.reset_mock()

        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete', text='NEW')))
        self.advance_time_and_run(1)

        # High score 1

        # award slide is 4 secs, but we only advanced 1, so the next request
        # should not have been sent yet
        self.assertEqual(1, self._events['high_score_enter_initials'])

        # advance 4 secs and it should be sent
        self.advance_time_and_run(4)
        self.assertEqual(2, self._events['high_score_enter_initials'])

        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete', text='P2')))
        self.advance_time_and_run(5)

        # High score done

        self.assertFalse(self.machine.modes["high_score"].active)

        # verify the data is accurate
        new_score_data = OrderedDict()
        new_score_data['score'] = [('NEW', 10000000),
                                   ('P2', 8000000),
                                   ('BRI', 7050550),
                                   ('GHK', 93060),
                                   ('JK', 87890)]
        new_score_data['loops'] = []

        self.assertEqual(new_score_data,
                         self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    def test_2_high_scores_and_timeout(self):
        self.mock_event("high_score_enter_initials")
        self.mock_event("high_score_award_display")
        self.machine.modes["high_score"].high_scores = OrderedDict(
            score=[('BRI', 7050550),
                   ('GHK', 93060),
                   ('JK', 87890),
                   ('QC', 87890),
                   ('MPF', 1000)])

        self.start_game(2)
        self.machine.game.player_list[0].score = 8000000
        self.machine.game.player_list[1].score = 10000000
        self._bcp_client.send.reset_mock()
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)

        # GC
        self.assertEventCalledWith('high_score_enter_initials', award='GRAND CHAMPION', player_num=2, value=10000000)
        self.mock_event("high_score_enter_initials")
        # wait for timeout
        self.advance_time_and_run(20)
        self.assertEventNotCalled("high_score_award_display")

        # Player 2 did not react. Player 1 will be GC
        # no further slide in between
        self.assertEventCalledWith('high_score_enter_initials', award='GRAND CHAMPION', player_num=1, value=8000000)

        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete', text='P2')))
        self.advance_time_and_run(.5)
        self.assertEventCalledWith("high_score_award_display",
                                   award='GRAND CHAMPION', player_name='P2', value=8000000)
        self.mock_event("high_score_award_display")
        self.advance_time_and_run(4)

        # High score done

        self.assertFalse(self.machine.modes["high_score"].active)

        # verify the data is accurate
        new_score_data = OrderedDict()
        new_score_data['score'] = [('P2', 8000000),
                                   ('BRI', 7050550),
                                   ('GHK', 93060),
                                   ('JK', 87890),
                                   ('QC', 87890)]
        new_score_data['loops'] = []

        self.assertEqual(new_score_data,
                         self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    def test_new_score_to_incomplete_list(self):
        self.mock_event("high_score_enter_initials")
        self.machine.modes["high_score"].high_scores = OrderedDict(
            score=[('BRI', 7050550),
                   ('GHK', 93060),
                   ('MPF', 1000)])

        self.start_game(4)
        self.machine.game.player_list[0].score = 1500
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)

        self.assertEqual(1, self._events['high_score_enter_initials'])

        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete', text='NEW')))

        self.advance_time_and_run(5)
        self.assertFalse(self.machine.modes["high_score"].active)

        new_score_data = OrderedDict()
        new_score_data['score'] = [('BRI', 7050550),
                                   ('GHK', 93060),
                                   ('NEW', 1500),
                                   ('MPF', 1000)]
        new_score_data['loops'] = []

        self.assertEqual(new_score_data,
                         self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    def test_more_than_3_chars(self):
        self.mock_event("high_score_enter_initials")
        self.machine.modes["high_score"].high_scores = OrderedDict(
            score=[('BRI', 7050550),
                   ('GHK', 93060),
                   ('JK', 87890),
                   ('QC', 87890),
                   ('MPF', 1000)])

        self.start_game(4)
        self.machine.game.player_list[0].score = 8000000
        self.machine.game.player_list[1].score = 1000
        self.machine.game.player_list[2].score = 1000
        self.machine.game.player_list[3].score = 1000
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)

        self.assertEqual(1, self._events['high_score_enter_initials'])

        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete',
                                                                   text='NEWNEW')))

        self.advance_time_and_run(5)
        self.assertFalse(self.machine.modes["high_score"].active)

        new_score_data = OrderedDict()
        new_score_data['score'] = [('NEWNEW', 8000000),
                                   ('BRI', 7050550),
                                   ('GHK', 93060),
                                   ('JK', 87890),
                                   ('QC', 87890)]
        new_score_data['loops'] = []

        self.assertEqual(new_score_data,
                         self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    def test_multiple_awards(self):
        self.mock_event("high_score_enter_initials")
        self.mock_event("high_score_award_display")
        # tests multiple awards (score & loops)
        # also tests 2 players getting an award for one slot, so only the
        # highest one should be presented
        # also tests the order (score first, then loops)
        self.machine.modes["high_score"].high_scores = OrderedDict()
        self.machine.modes["high_score"].high_scores['score'] = [('BRI', 7050550),
                                                              ('GHK', 93060),
                                                              ('JK', 87890),
                                                              ('QC', 87890),
                                                              ('MPF', 1000)]
        self.machine.modes["high_score"].high_scores['loops'] = [('BIL', 2)]

        self.start_game(4)
        self.machine.game.player_list[0].score = 8000000
        self.machine.game.player_list[1].score = 10000000
        self.machine.game.player_list[2].score = 1000
        self.machine.game.player_list[3].score = 1000
        self.machine.game.player_list[0].loops = 50
        self.machine.game.player_list[1].loops = 4
        self.machine.game.player_list[2].loops = 1
        self.machine.game.player_list[3].loops = 0
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes["high_score"].active)

        # GC

        self.assertEqual(1, self._events['high_score_enter_initials'])
        self._bcp_client.send.reset_mock()

        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete', text='NEW')))
        self.advance_time_and_run(.5)
        self.assertEventCalledWith("high_score_award_display",
                                   award='GRAND CHAMPION', player_name='NEW', value=10000000)
        self.mock_event("high_score_award_display")
        self.advance_time_and_run(4)

        # High score 1

        self.assertEqual(2, self._events['high_score_enter_initials'])

        self._bcp_client.receive_queue.put_nowait(('trigger', dict(name='text_input_high_score_complete', text='P1')))
        self.advance_time_and_run(.5)
        self.assertEventCalledWith("high_score_award_display",
                                   award='HIGH SCORE 1', player_name='P1', value=8000000)
        self.mock_event("high_score_award_display")
        self.advance_time_and_run(4)

        # Loops champ should not ask again but show a slide
        self.assertTrue(self.machine.modes["high_score"].active)
        self.advance_time_and_run(.5)
        self.assertEventCalledWith("high_score_award_display",
                                   award='LOOP CHAMP', player_name='P1', value=50)
        self.mock_event("high_score_award_display")
        self.advance_time_and_run(4)

        # High score done
        self.assertFalse(self.machine.modes["high_score"].active)

        # verify the data is accurate

        new_score_data = OrderedDict()
        new_score_data['score'] = [('NEW', 10000000),
                                   ('P1', 8000000),
                                   ('BRI', 7050550),
                                   ('GHK', 93060),
                                   ('JK', 87890)]
        new_score_data['loops'] = [('P1', 50)]

        # only ask every player once
        self.assertEqual(2, self._events['high_score_enter_initials'])

        self.assertEqual(new_score_data,
                         self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    def test_score_from_nonexistent_player_var(self):
        self.machine.modes["high_score"].high_scores = OrderedDict(
            score=[('BRI', 7050550),
                   ('GHK', 93060),
                   ('JK', 87890),
                   ('QC', 87890),
                   ('MPF', 1000)])

        self.start_game(4)
        self.machine.game.end_game()
        self.advance_time_and_run()

        self.assertFalse(self.machine.modes["high_score"].active)

        new_score_data = OrderedDict()
        new_score_data['score'] = [('BRI', 7050550),
                                   ('GHK', 93060),
                                   ('JK', 87890),
                                   ('QC', 87890),
                                   ('MPF', 1000)]
        new_score_data['loops'] = []

        self.assertEqual(new_score_data,
                         self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    def _get_mock_data(self):
        if self._testMethodName == "testInvalidData":
            new_score_data = OrderedDict()
            new_score_data['score'] = [(13, '123'),
                                       (123.1, 4),
                                       (1)]
            new_score_data['loops'] = ""

            return {"high_scores": new_score_data}
        elif self._testMethodName == "testLoadData":
            new_score_data = OrderedDict()
            new_score_data['score'] = [('BRI', 5),
                                       ('GHK', 4),
                                       ('JK', 3),
                                       ('QC', 2),
                                       ('MPF', 1)]
            new_score_data['loops'] = [('MPF', 7)]
            return {"high_scores": new_score_data}
        elif self._testMethodName == "test_reset":
            new_score_data = OrderedDict(
                score=[('ABC', 1000000),
                       ('ABC', 900000),
                       ('ABC', 800000),
                       ('ABC', 700000),
                       ('ABC', 600000)],
                loops=[('MPF', 500000)])
            return {"high_scores": new_score_data}

        return super()._get_mock_data()

    def testLoadData(self):
        new_score_data = OrderedDict()
        new_score_data['score'] = [('BRI', 5),
                                   ('GHK', 4),
                                   ('JK', 3),
                                   ('QC', 2),
                                   ('MPF', 1)]
        new_score_data['loops'] = [('MPF', 7)]
        self.assertEqual(new_score_data, self.machine.modes["high_score"].high_scores)
        # no changes yet
        self.assertEqual(None, self.machine.modes["high_score"].data_manager.written_data)

    def testInvalidData(self):
        self.start_game(4)
        self.machine.game.game_ending()
        self.advance_time_and_run(10)

        # this should reload the defaults
        new_score_data = OrderedDict()
        new_score_data['score'] = [('BRI', 4242),
                                   ('GHK', 2323),
                                   ('JK', 1337),
                                   ('QC', 42),
                                   ('MPF', 23)]
        new_score_data['loops'] = [('JK', 42)]

        self.assertEqual(new_score_data, self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)

    @test_config_directory("tests/machine_files/high_score_reverse/")
    def test_reverse_sort(self):
        self.start_game(4)
        self.machine.game.game_ending()
        self.advance_time_and_run()

        new_score_data = OrderedDict()
        new_score_data['score'] = [('BRI', 4242),
                                   ('GHK', 2323),
                                   ('JK', 1337),
                                   ('QC', 42),
                                   ('MPF', 23)]
        new_score_data['loops'] = [('JK', 42)]
        new_score_data['time_to_wizard'] = [('JK', 300),
                                            ('BM', 350)]
        self.assertEqual(new_score_data, self.machine.modes["high_score"].high_scores)
        self.assertEqual(new_score_data, self.machine.modes["high_score"].data_manager.written_data)



"""Test MyPinballs Platform."""
import time

from mpf.platforms.interfaces.segment_display_platform_interface import FlashingType
from mpf.tests.MpfTestCase import MpfTestCase
from mpf.tests.loop import MockSerial


class MockMypinballsSocket(MockSerial):

    """Serial mock."""

    def read(self, length):
        """Read from serial."""
        del length
        if not self.queue:
            return b''
        msg = self.queue.pop()

        return msg

    def read_ready(self):
        """True if ready to read."""
        return bool(self.queue)

    def write_ready(self):
        """True if ready to write."""
        return True

    def write(self, msg):
        """Write message."""
        parts = msg.split(b'\n')

        # remove last newline
        assert parts.pop() == b''

        for part in parts:
            self._handle_msg(part)

        return len(msg)

    def _handle_msg(self, msg):
        if msg in self.permanent_commands and msg not in self.expected_commands:
            self.queue.append(self.permanent_commands[msg])
            return len(msg)

        # print("Serial received: " + "".join("\\x%02x" % b for b in msg) + " len: " + str(len(msg)))
        if msg not in self.expected_commands:
            self.crashed = True
            raise AssertionError("Unexpected command: " + msg.decode() + "(" + "".join("\\x%02x" % b for b in msg) +
                                 ") len: " + str(len(msg)))

        if self.expected_commands[msg] is not False:
            self.queue.append(self.expected_commands[msg])

        del self.expected_commands[msg]

        return len(msg)

    def __init__(self):
        super().__init__()
        self.name = "SerialMock"
        self.expected_commands = {}
        self.queue = []
        self.permanent_commands = {}
        self.crashed = False


class MyPinballsPlatformTest(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/mypinballs/'

    def _mock_loop(self):
        self.clock.mock_serial("/dev/ttyUSB0", self.serialMock)

    def tearDown(self):
        self.assertFalse(self.serialMock.crashed)
        super().tearDown()

    def get_platform(self):
        return False

    def _wait_for_processing(self):
        start = time.time()
        while self.serialMock.expected_commands and not self.serialMock.crashed and time.time() < start + 10:
            self.advance_time_and_run(.01)

    def setUp(self):
        self.serialMock = MockMypinballsSocket()

        # all display are reset at startup
        self.serialMock.expected_commands = {
            b'3:1': False,
            b'3:2': False,
            b'3:6': False,
        }
        self.serialMock.permanent_commands = {}
        super().setUp()

        self._wait_for_processing()

        self.assertFalse(self.serialMock.expected_commands)

    def testPlatform(self):
        self.serialMock.expected_commands = {
            b'1:1:1234': False,
        }
        self.machine.segment_displays["display1"].add_text("1234", key="score")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # change text (with same key)
        self.serialMock.expected_commands = {
            b'1:1:1337': False,
        }
        self.machine.segment_displays["display1"].add_text("1337", key="score")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # change text (with same key)
        self.serialMock.expected_commands = {
            b'1:1:42?23': False,
        }
        self.machine.segment_displays["display1"].add_text("42 23", key="score")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        # set to empty
        self.serialMock.expected_commands = {
            b'3:1': False,
        }
        self.machine.segment_displays["display1"].remove_text_by_key("score")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'1:2:424242': False,
        }
        self.machine.segment_displays["display2"].add_text("424242")
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'2:2:424242': False,
        }
        self.machine.segment_displays["display2"].set_flashing(FlashingType.FLASH_ALL)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'1:2:424242': False,
        }
        self.machine.segment_displays["display2"].set_flashing(FlashingType.NO_FLASH)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)

        self.serialMock.expected_commands = {
            b'4:2:424242': False,
        }
        self.machine.segment_displays["display2"].set_flashing(FlashingType.FLASH_MATCH)
        self._wait_for_processing()
        self.assertFalse(self.serialMock.expected_commands)



"""A direct light on a fast controller."""
import logging

from mpf.core.utility_functions import Util
from mpf.platforms.interfaces.light_platform_interface import LightPlatformSoftwareFade


class FASTMatrixLight(LightPlatformSoftwareFade):

    """A direct light on a fast controller."""

    __slots__ = ["log", "number", "send", "platform"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, number, sender, machine, fade_interval_ms: int, platform) -> None:
        """Initialise light."""
        super().__init__(number, machine.clock.loop, fade_interval_ms)
        self.log = logging.getLogger('FASTMatrixLight')
        self.send = sender
        self.platform = platform

    def set_brightness(self, brightness: float):
        """Set matrix light brightness."""
        self.send('L1:{},{}'.format(self.number, Util.int_to_hex_string(int(brightness * 255))))

    def get_board_name(self):
        """Return the board of this light."""
        if self.platform.machine_type == 'wpc':
            return "FAST WPC"

        coil_index = 0
        number = Util.hex_string_to_int(self.number)
        for board_obj in self.platform.io_boards.values():
            if coil_index <= number < coil_index + board_obj.driver_count:
                return "FAST Board {}".format(str(board_obj.node_id))
            coil_index += board_obj.driver_count

        # fall back if not found
        return "FAST Unknown Board"

    def is_successor_of(self, other):
        """Return true if the other light has the previous number."""
        raise AssertionError("Not possible in FASTMatrix.")

    def get_successor_number(self):
        """Return next number."""
        raise AssertionError("Not possible in FASTMatrix.")

    def __lt__(self, other):
        """Order lights by string."""
        return self.number < other.number



from mpf.devices.light import Light
from mpf.tests.MpfTestCase import MpfTestCase


class TestDeviceCollection(MpfTestCase):
    def get_config_file(self):
        return 'test_device_collection.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/device_collection/'

    def test_accessing_devices(self):

        led1 = self.machine.lights['led1']
        led2 = self.machine.lights['led2']
        led3 = self.machine.lights['led3']
        led4 = self.machine.lights['led4']

        self.assertTrue(isinstance(led1, Light))
        self.assertTrue(isinstance(led2, Light))
        self.assertTrue(isinstance(led3, Light))
        self.assertTrue(isinstance(led4, Light))

        self.assertEqual(self.machine.lights["led1"], led1)
        self.assertEqual(self.machine.lights["led2"], led2)
        self.assertEqual(self.machine.lights["led3"], led3)
        self.assertEqual(self.machine.lights["led4"], led4)

    def test_tags(self):
        led1 = self.machine.lights['led1']
        led2 = self.machine.lights['led2']
        led3 = self.machine.lights['led3']
        led4 = self.machine.lights['led4']

        self.assertEqual([], self.machine.lights.items_tagged('fake_tag'))

        self.assertIn(led1, self.machine.lights.items_tagged('tag1'))
        self.assertIn(led2, self.machine.lights.items_tagged('tag1'))
        self.assertNotIn(led3, self.machine.lights.items_tagged('tag1'))
        self.assertNotIn(led4, self.machine.lights.items_tagged('tag1'))



"""Test accelerometer device."""
import math

from mpf.tests.MpfTestCase import MpfTestCase


class TestAccelerometer(MpfTestCase):

    def __init__(self, methodName):
        super().__init__(methodName)
        # this is the first test. give it some more time
        self.expected_duration = 2

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/accelerometer/'

    def _event_level1(self, **kwargs):
        del kwargs
        self._level1 = True

    def _event_level2(self, **kwargs):
        del kwargs
        self._level2 = True

    def test_leveling_different_base_angle(self):
        accelerometer = self.machine.accelerometers["test_accelerometer"]

        # change base angle to 6.5 degree
        accelerometer.config['level_x'] = math.sin(math.radians(6.5))
        accelerometer.config['level_y'] = 0
        accelerometer.config['level_z'] = math.cos(math.radians(6.5))

        # machine should be 6.5 degree off
        accelerometer.update_acceleration(0.0, 0.0, 1)
        self.assertAlmostEqual(math.radians(6.5), accelerometer.get_level_xz())
        self.assertAlmostEqual(0.0, accelerometer.get_level_yz())
        self.assertAlmostEqual(math.radians(6.5), accelerometer.get_level_xyz())

        # add a small tilt
        accelerometer.update_acceleration(0.0, math.sin(math.radians(5)), math.cos(math.radians(5)))
        self.assertAlmostEqual(math.radians(6.5), accelerometer.get_level_xz())
        self.assertAlmostEqual(math.radians(5), accelerometer.get_level_yz())

        # leveled
        accelerometer.update_acceleration(math.sin(math.radians(6.5)), 0.0, math.cos(math.radians(6.5)))
        self.assertAlmostEqual(0.0, accelerometer.get_level_xz())
        self.assertAlmostEqual(0.0, accelerometer.get_level_yz())
        self.assertAlmostEqual(0.0, accelerometer.get_level_xyz())

        # leveled + tilt
        accelerometer.update_acceleration(math.sin(math.radians(6.5)) / math.cos(math.radians(6.5)),
                                          math.sin(math.radians(5)) / math.cos(math.radians(5)),
                                          1)
        self.assertAlmostEqual(0.0, accelerometer.get_level_xz())
        self.assertAlmostEqual(math.radians(5), accelerometer.get_level_yz())

    def test_leveling(self):
        accelerometer = self.machine.accelerometers["test_accelerometer"]

        self._level1 = False
        self._level2 = False
        self.machine.events.add_handler("event_level1", self._event_level1)
        self.machine.events.add_handler("event_level2", self._event_level2)

        # perfectly leveled
        accelerometer.update_acceleration(0.0, 0.0, 1.0)
        self.assertAlmostEqual(0.0, accelerometer.get_level_xz())
        self.assertAlmostEqual(0.0, accelerometer.get_level_yz())
        self.assertAlmostEqual(0.0, accelerometer.get_level_xyz())
        self.machine_run()
        self.assertFalse(self._level1)
        self.assertFalse(self._level2)

        # 90 degree on the side
        for _ in range(100):
            accelerometer.update_acceleration(1.0, 0.0, 0.0)
        self.assertAlmostEqual(math.pi / 2, accelerometer.get_level_xz())
        self.assertAlmostEqual(0, accelerometer.get_level_yz())
        self.assertAlmostEqual(math.pi / 2, accelerometer.get_level_xyz())
        self.machine_run()
        self.assertTrue(self._level1)
        self.assertTrue(self._level2)

        # 90 degree on the back
        accelerometer.update_acceleration(0.0, 1.0, 0.0)
        self.machine_run()
        self.assertAlmostEqual(0, accelerometer.get_level_xz())
        self.assertAlmostEqual(math.pi / 2, accelerometer.get_level_yz())
        self.assertAlmostEqual(math.pi / 2, accelerometer.get_level_xyz())

        # 45 degree on the side
        accelerometer.update_acceleration(0.5, 0.0, 0.5)
        self.machine_run()
        self.assertAlmostEqual(math.pi / 4, accelerometer.get_level_xz())
        self.assertAlmostEqual(0, accelerometer.get_level_yz())
        self.assertAlmostEqual(math.pi / 4, accelerometer.get_level_xyz())

        # 3.01 degree
        self._level1 = False
        self._level2 = False
        for _ in range(100):
            accelerometer.update_acceleration(0.0, 0.05, 0.95)
        self.machine_run()
        self.assertTrue(self._level1)
        self.assertFalse(self._level2)

        # 6.34 degree
        self._level1 = False
        self._level2 = False
        for _ in range(100):
            accelerometer.update_acceleration(0.0, 0.1, 0.9)
        self.machine_run()
        self.assertTrue(self._level1)
        self.assertTrue(self._level2)

    def _event_hit1(self, **kwargs):
        del kwargs
        self._hit1 = True

    def _event_hit2(self, **kwargs):
        del kwargs
        self._hit2 = True

    def test_hits(self):
        accelerometer = self.machine.accelerometers["test_accelerometer"]

        # perfectly leveled
        accelerometer.update_acceleration(0.0, 0.0, 1.0)
        self.machine_run()

        self._hit1 = False
        self._hit2 = False
        self.machine.events.add_handler("event_hit1", self._event_hit1)
        self.machine.events.add_handler("event_hit2", self._event_hit2)

        # some noise
        accelerometer.update_acceleration(0.01, 0.05, 0.99)
        self.machine_run()
        self.assertFalse(self._hit1)
        self.assertFalse(self._hit2)

        # hit from the side
        accelerometer.update_acceleration(0.4, 0.4, 1.0)
        self.machine_run()
        self.assertTrue(self._hit1)
        self.assertFalse(self._hit2)
        self._hit1 = False
        self._hit2 = False

        # and it calms
        accelerometer.update_acceleration(0.01, 0.05, 0.99)
        accelerometer.update_acceleration(0.01, 0.05, 0.99)
        accelerometer.update_acceleration(0.01, 0.05, 0.99)
        accelerometer.update_acceleration(0.01, 0.05, 0.99)
        accelerometer.update_acceleration(0.01, 0.05, 0.99)
        self.machine_run()
        self.assertFalse(self._hit1)
        self.assertFalse(self._hit2)

        # strong hit from the side
        accelerometer.update_acceleration(1.5, 0.4, 1.0)
        self.machine_run()
        self.assertTrue(self._hit1)
        self.assertTrue(self._hit2)



"""Base class for MPF controllers."""
import abc

from mpf.core.logging import LogMixin
MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class MpfController(LogMixin, metaclass=abc.ABCMeta):

    """Base class for MPF controllers."""

    __slots__ = ["machine"]

    module_name = None  # type: str
    config_name = None  # type: str

    def __init__(self, machine: "MachineController") -> None:
        """Initialise controller.

        Args:
        ----
            machine(mpf.core.machine.MachineController): the machine controller
        """
        super().__init__()
        self.machine = machine  # type: MachineController

        if not self.config_name:
            raise AssertionError("Please specify a config name for {}".format(self))

        self.configure_logging(
            self.module_name if self.module_name else self.__class__.__name__,
            self.machine.config['logging']['console'][self.config_name],
            self.machine.config['logging']['file'][self.config_name])

        self.debug_log("Loading the {}".format(self.module_name))



"""I2C platform which uses the smbus interface on linux via the smbus2 python extension."""
import asyncio
from typing import Tuple

import logging

from mpf.platforms.interfaces.i2c_platform_interface import I2cPlatformInterface

from mpf.core.platform import I2cPlatform

EXTENSION_LOADED = True
# pylint: disable-msg=ungrouped-imports
try:
    from smbus2_asyncio import SMBus2Asyncio
except ImportError:     # pragma: no cover
    SMBus2Asyncio = None


class Smbus2I2cDevice(I2cPlatformInterface):

    """A i2c device on smbus2."""

    __slots__ = ["loop", "platform", "busses", "address", "smbus"]

    # noinspection PyNoneFunctionAssignment
    def __init__(self, number: str, platform, busses) -> None:
        """Initialise smbus2 device."""
        super().__init__(number)
        self.loop = platform.machine.clock.loop
        self.platform = platform
        self.busses = busses
        bus, self.address = self._get_i2c_bus_address(number)
        self.smbus = self._get_i2c_bus(bus)

    async def open(self):
        """Open device (if not already open)."""
        while not self.smbus.smbus:
            try:
                await self.smbus.open()
            except FileNotFoundError:
                if not self.platform.machine.options["production"]:
                    raise

                # if we are in production mode retry
                await asyncio.sleep(.1)
                self.platform.log.debug("Connection to %s failed. Will retry.", self.number)
            else:
                break

    @staticmethod
    def _get_i2c_bus_address(address) -> Tuple[int, int]:
        """Split and return bus + address."""
        if isinstance(address, int):
            return 0, address
        bus, address = address.split("-", 1)
        return bus, int(address)

    # noinspection PyCallingNonCallable
    def _get_i2c_bus(self, bus) -> SMBus2Asyncio:
        """Get or open handle for i2c bus."""
        if bus in self.busses:
            return self.busses[bus]
        handle = SMBus2Asyncio(bus)
        self.busses[bus] = handle
        return handle

    async def i2c_read8(self, register):
        """Read a byte from I2C."""
        return await self.smbus.read_byte_data(self.address, int(register))

    def i2c_write8(self, register, value):
        """Write a byte to I2C."""
        asyncio.ensure_future(self.smbus.write_byte_data(self.address, int(register), int(value)))
        # this does not return

    async def i2c_read_block(self, register, count):
        """Read a block from I2C."""
        return await self.smbus.read_i2c_block_data(self.address, int(register), int(count))


class Smbus2(I2cPlatform):

    """I2C platform which uses the smbus interface on linux via the smbus2 python extension."""

    __slots__ = ["_i2c_busses"]

    def __init__(self, machine):
        """Initialise Smbus2 platform."""
        super().__init__(machine)
        self._i2c_busses = {}
        self.log = logging.getLogger('Smbus2')

    async def initialize(self):
        """Check if smbus2 extension has been imported."""
        if not SMBus2Asyncio:
            raise AssertionError("smbus2 python extension missing. Please run: pip3 install smbus2_asyncio")

    async def configure_i2c(self, number: str) -> Smbus2I2cDevice:
        """Configure device on smbus2."""
        device = Smbus2I2cDevice(number, self, self._i2c_busses)
        await device.open()
        return device



from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestTooLongExitCountDelay(MpfTestCase):

    def get_config_file(self):
        return 'test_too_long_exit_count_delay.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def _ball_drained(self, **kwargs):
        del kwargs
        self._num_balls_drained += 1

    def put_four_balls_in_trough(self):
        self.machine.switch_controller.process_switch('s_trough_1', 1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.advance_time_and_run(1)
        self.assertEqual(4, self.machine.ball_devices["trough"].balls)
        self.assertEqual(0, self.machine.playfield.balls)

        self.trough_coil = self.machine.coils["trough_eject"]
        self.plunger_coil = self.machine.coils["plunger_eject"]

        self.trough_coil.pulse = MagicMock()
        self.plunger_coil.pulse = MagicMock()

        self._num_balls_drained = 0
        self.machine.events.add_handler('ball_drained', self._ball_drained)

    def test_eject_to_plunger(self):
        # tests eject from trough to plunger where the ball enters the plunger
        # before the exit_count_delay of the trough

        self.put_four_balls_in_trough()

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # default pulse
        self.assertTrue(self.trough_coil.pulse.called)
        self.trough_coil.pulse = MagicMock()

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(.1)

        # ball goes into plunger
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 1)
        assert not self.plunger_coil.pulse.called
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("ejecting", self.machine.ball_devices["plunger"]._state)

        # player hits the launch button
        self.machine.switch_controller.process_switch('s_launch', 1)
        self.machine.switch_controller.process_switch('s_launch', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_plunger', 0)
        self.advance_time_and_run(1)

        # ball moves from plunger lane to playfield
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.machine.switch_controller.process_switch('s_playfield', 0)
        self.advance_time_and_run(.1)

        self.assertTrue(self.plunger_coil.pulse.called)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 0)
        self.assertEqual(self.machine.ball_devices["playfield"].balls, 1)
        self.assertEqual("idle", self.machine.ball_devices["trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["plunger"]._state)
        self.assertEqual(4, self.machine.ball_controller.num_balls_known)

    def test_ball_fell_back_in_trough_before_exit_count_delay(self):
        self.put_four_balls_in_trough()

        self.machine.playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # default pulse
        self.assertTrue(self.trough_coil.pulse.called)
        self.trough_coil.pulse = MagicMock()

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.machine.switch_controller.process_switch('s_trough_2', 0)
        self.machine.switch_controller.process_switch('s_trough_3', 0)
        self.machine.switch_controller.process_switch('s_trough_4', 0)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)
        self.machine.switch_controller.process_switch('s_trough_3', 1)
        self.machine.switch_controller.process_switch('s_trough_4', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.machine.switch_controller.process_switch('s_trough_jam', 0)
        self.advance_time_and_run(.2)

        # ball falls back in trough
        self.machine.switch_controller.process_switch('s_trough_jam', 1)
        self.advance_time_and_run(2)

        self.assertEqual(self.machine.ball_devices["trough"].balls, 3)
        self.assertEqual(self.machine.ball_devices["plunger"].balls, 0)
        self.assertEqual(self.machine.ball_devices["playfield"].balls, 0)

        # after the eject timeout, the trough will realize it has a ball
        self.advance_time_and_run(10)
        self.assertEqual(self.machine.ball_devices["trough"].balls, 4)



"""Command to format yaml files."""
import sys

import os
from mpf.commands import MpfCommandLineParser

try:
    from prompt_toolkit.formatted_text import HTML
    from prompt_toolkit.shortcuts import message_dialog
    from prompt_toolkit.styles import Style
    from prompt_toolkit.shortcuts import button_dialog
    from prompt_toolkit.shortcuts import input_dialog
except ImportError:
    HTML = None
    message_dialog = None
    Style = None
    button_dialog = None
    input_dialog = None


SUBCOMMAND = True


class Command(MpfCommandLineParser):

    """Run a text unit test from cli."""

    def __init__(self, args, path):
        """Parse args."""
        super().__init__(args, path)
        self.current_path = path
        if self.in_machine_folder():
            self.machine_path = self.current_path

        if not HTML:
            raise AssertionError("You need to install the [cli] or [all] feature of mpf to use this.")

        self.example_style = Style.from_dict({
            'dialog': 'bg:#f2521d',
            'dialog frame.label': 'bg:#ffffff #000000',
            'dialog.body': 'bg:#000000 #FFFFFF',
            'dialog shadow': 'bg:#79290E',
            'dialog text-area': 'bg:#f2521d #FFFFFF'
        })

        self.creation_loop()

    def creation_loop(self):
        """Set up the main creation loop."""
        selection = button_dialog(
            title='Selection',
            text='What would you like to create?',

            buttons=[
                ('Machine', 'machine'),
                ('Mode', 'mode'),
                ('Show', 'show'),
                ('Cancel', 'cancel'),
            ],
            style=self.example_style
        ).run()
        if selection == 'machine':
            self.create_machine_config()
        elif selection == 'mode':
            if self.in_machine_folder():
                self.create_mode()
            else:
                self.show_not_in_machine_folder_dialog()
        elif selection == 'show':
            if self.in_machine_folder():
                self.create_show()
            else:
                self.show_not_in_machine_folder_dialog()
        else:
            sys.exit()
        self.creation_loop()

    def create_mode(self):
        """Create a mode."""
        mode_name = input_dialog(
            title='Mode',
            text='Cool, got a name for your mode?',
            style=self.example_style).run()

        # create mode folder
        mode_path = os.path.join(self.machine_path, "modes", mode_name)
        if not os.path.exists(mode_path):
            self.create_mode_structure(mode_name, mode_path, self.machine_path)
            message_dialog(
                title='Mode',
                text=HTML(
                    '''<style fg="green">Success:</style> Created mode {}.
                    \nDon\'t forget to add it to your mode list.'''.format(
                        mode_name)),
                style=self.example_style
            ).run()

        else:
            message_dialog(
                title='Mode',
                text=HTML(
                    '<style fg="red">Error:</style> A mode with this name exists already\nPlease pick another one'),
                style=self.example_style
            ).run()
            self.create_mode()

    def create_machine_config(self):
        """Create a machine config."""
        config_name = input_dialog(
            title='Machine Config',
            text='Please enter the name of your machine config:',
            style=self.example_style).run()

        if config_name is None:
            sys.exit()

        # create machine_config folder
        self.machine_path = os.path.join(self.current_path, config_name)

        if not os.path.exists(self.machine_path):

            self.create_machine_config_structure(config_name, self.machine_path)
            self.current_path = self.machine_path
            message_dialog(
                title='Mode',
                text=HTML(
                    '''<style fg="green">Success:</style> Created machine config {}.
                    \nYou can now create modes and shows.'''.format(
                        config_name)),
                style=self.example_style
            ).run()
        else:
            message_dialog(
                title='Mode',
                text=HTML(
                    '''<style fg="red">Error:</style> A machine config with this name exists already
                    \nPlease pick another one')'''),
                style=self.example_style
            ).run()
            self._create_machine_config()

    def create_show(self):
        """Create a show."""
        show_name = input_dialog(
            title='Mode',
            text='Cool, got a name for your show?',
            style=self.example_style).run()

        # create shows folder
        show_path = os.path.normpath(os.path.join(self.machine_path, "shows", show_name))
        shows_dir = os.path.normpath(os.path.join(self.machine_path, "shows"))

        if not os.path.exists(show_path):
            if self.in_machine_folder():
                self.create_show_structure(show_name, shows_dir, self.machine_path)
                message_dialog(
                    title='Shows',
                    text=HTML('<style fg="green">Success:</style> Created show {}.'.format(show_name)),
                    # text='Success: Created machine config {}.'.format(config_name),
                    style=self.example_style
                ).run()
            else:
                self.show_not_in_machine_folder_dialog()

        else:
            message_dialog(
                title='Mode',
                text=HTML(
                    '<style fg="red">Error:</style> A show with this name already exists\nPlease pick another one!'),
                style=self.example_style
            ).run()
            self._create_machine_config()

    def in_machine_folder(self):
        """Check if current directory is a machine folder."""
        return os.path.exists(os.path.join(self.current_path, "config"))

    def show_not_in_machine_folder_dialog(self):
        """Show error message if current directory is not a machine folder."""
        message_dialog(
            title='Wrong directory',
            text=HTML(
                '<style fg="red">Error:</style> You dont\'t seem to be in a machine folder.\nPlease switch to one!'),
            style=self.example_style
        ).run()
        sys.exit()

    @staticmethod
    def create_machine_config_structure(config_name, machine_path):
        """Create the basic hierarchy of a new machine folder including test and config files."""
        os.makedirs(os.path.normpath(machine_path))
        os.makedirs(os.path.normpath(os.path.join(machine_path, "config")))
        with open(os.path.normpath(os.path.join(machine_path, "config", "config.yaml")), "w") as f:
            f.write("#config_version=5")

        # create test folder
        os.makedirs(os.path.normpath(os.path.join(machine_path, "tests")))

        test_file_name = "test_{}.yaml".format(config_name)
        with open(os.path.normpath(os.path.join(machine_path, "tests", test_file_name)), "w"):
            pass

    @staticmethod
    def create_show_structure(show_name, shows_dir, machine_path):
        """Create the basic hierarchy of a new show folder."""
        if not os.path.exists(shows_dir):
            os.makedirs(shows_dir)
        file_content = "#show_version=5"
        show_file_name = "{}.yaml".format(show_name)
        config_path = os.path.normpath(os.path.join(machine_path, "shows", show_file_name))
        Command.write_config_file(config_path, file_content)

    @staticmethod
    def create_mode_structure(mode_name, mode_path, machine_path):
        """Create the basic hierarchy of a new mode folder."""
        os.makedirs(mode_path)
        # create config folder
        os.makedirs(os.path.normpath(os.path.join(machine_path, "modes", mode_name, "config")))
        # create config file
        file_content = """#config_version=5

            mode:
                start_events: ball_started
                priority: 100
            """
        config_file_name = "{}.yaml".format(mode_name)
        config_path = os.path.normpath(os.path.join(machine_path, "modes", mode_name, "config", config_file_name))
        Command.write_config_file(
            config_path, file_content)

    @staticmethod
    def write_config_file(path, config_file_content):
        """Create config files for machine folders or modes."""
        # create config file
        with open(path, "w") as f:
            f.write(config_file_content)



"""Logic Blocks devices."""
from random import shuffle

from typing import Any, List, Optional

from mpf.core.delays import DelayManager
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler
from mpf.core.machine import MachineController
from mpf.core.mode import Mode
from mpf.core.mode_device import ModeDevice
from mpf.core.player import Player
from mpf.core.system_wide_device import SystemWideDevice
from mpf.core.utility_functions import Util


class LogicBlockState:

    """Represents the state of a logic_block."""

    __slots__ = ["enabled", "completed", "value"]

    def __init__(self):
        """Initialise state."""
        self.enabled = False
        self.completed = False
        self.value = None


@DeviceMonitor("value", "enabled", "completed")
class LogicBlock(SystemWideDevice, ModeDevice):

    """Parent class for each of the logic block classes."""

    __slots__ = ["_state", "_start_enabled", "player_state_variable"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Initialize logic block."""
        super().__init__(machine, name)
        self._state = None          # type: Optional[LogicBlockState]
        self._start_enabled = None  # type: Optional[bool]

        self.player_state_variable = "{}_state".format(self.name)
        '''player_var: (logic_block)_state

        desc: A dictionary that stores the internal state of the logic block
        with the name (logic_block). (In other words, a logic block called
        *mode1_hit_counter* will store its state in a player variable called
        ``mode1_hit_counter_state``).

        The state that's stored in this variable include whether the logic
        block is enabled and whether it's complete.
        '''

    async def _initialize(self):
        await super()._initialize()
        if self.config['start_enabled'] is not None:
            self._start_enabled = self.config['start_enabled']
        else:
            self._start_enabled = not self.config['enable_events']

    def add_control_events_in_mode(self, mode: Mode) -> None:
        """Do not auto enable this device in modes."""

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None) -> dict:
        """Validate logic block config."""
        del is_mode_config
        del debug_prefix
        if 'events_when_complete' not in config:
            config['events_when_complete'] = ['logicblock_' + self.name + '_complete']

        if 'events_when_hit' not in config:
            config['events_when_hit'] = ['logicblock_' + self.name + '_hit']

        self.machine.config_validator.validate_config(
            self.config_section, config, self.name, ("device", "logic_blocks_common"))

        self._configure_device_logging(config)
        return config

    @property
    def can_exist_outside_of_game(self) -> bool:
        """Return true if persist_state is not set."""
        return not bool(self.config['persist_state'])

    def get_start_value(self) -> Any:
        """Return the start value for this block."""
        raise NotImplementedError("implement")

    async def device_added_system_wide(self):
        """Initialise internal state."""
        self._state = LogicBlockState()
        self.value = self.get_start_value()
        await super().device_added_system_wide()
        if not self.config['enable_events']:
            self.enable()

        if self.config['persist_state']:
            self.raise_config_error("Cannot set persist_state for system-wide logic_blocks", 1)

        self.post_update_event()

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Restore internal state from player if persist_state is set or create new state."""
        super().device_loaded_in_mode(mode, player)
        if self.config['persist_state']:
            if not player.is_player_var(self.player_state_variable):
                player[self.player_state_variable] = LogicBlockState()
                # enable device ONLY when we create a new entry in the player
                if self._start_enabled:
                    mode.add_mode_event_handler("mode_{}_starting".format(mode.name),
                                                self.event_enable, priority=mode.priority + 1)

                self._state = player[self.player_state_variable]
                self.value = self.get_start_value()
            else:
                self._state = player[self.player_state_variable]
        else:
            self._state = LogicBlockState()
            self.value = self.get_start_value()
            if self._start_enabled:
                mode.add_mode_event_handler("mode_{}_starting".format(mode.name),
                                            self.event_enable, priority=mode.priority + 1)

        mode.add_mode_event_handler("mode_{}_starting".format(mode.name), self.post_update_event)

    def device_removed_from_mode(self, mode: Mode):
        """Unset internal state to prevent leakage."""
        super().device_removed_from_mode(mode)
        self._state = None

    @property
    def value(self):
        """Return value or None if that is currently not possible."""
        if self._state:
            return self._state.value

        return None

    @value.setter
    def value(self, value):
        """Set the value."""
        self._state.value = value

    @property
    def enabled(self):
        """Return if enabled."""
        return self._state and self._state.enabled

    @enabled.setter
    def enabled(self, value):
        """Set enable."""
        self._state.enabled = value

    @property
    def completed(self):
        """Return if completed."""
        return self._state and self._state.completed

    @completed.setter
    def completed(self, value):
        """Set if completed."""
        self._state.completed = value

    def post_update_event(self, **kwargs):
        """Post an event to notify about changes."""
        del kwargs
        value = self._state.value
        enabled = self._state.enabled
        self.machine.events.post("logicblock_{}_updated".format(self.name), value=value, enabled=enabled)
        '''event: logicblock_(name)_updated

        desc: The logic block called "name" has changed.

        This might happen when the block advanced, it was resetted or restored.

        args:
        value: The current value of this block.
        enabled: Whatever this block is enabled or not.
        '''

    def enable(self):
        """Enable this logic block.

        Automatically called when one of the
        enable_event events is posted. Can also manually be called.
        """
        super().enable()
        self.debug_log("Enabling")
        self.enabled = True
        self.post_update_event()

    def _post_hit_events(self, **kwargs):
        self.post_update_event()
        for event in self.config['events_when_hit']:
            self.machine.events.post(event, **kwargs)
            '''event: logicblock_(name)_hit

            desc: The logic block "name" was just hit.

            Note that this is the default hit event for logic blocks,
            but this can be changed in a logic block's "events_when_hit:"
            setting, so this might not be the actual event that's posted for
            all logic blocks in your machine.

            args: depend on the type
            '''

    @event_handler(0)
    def event_disable(self, **kwargs):
        """Event handler for disable event."""
        del kwargs
        self.disable()

    def disable(self):
        """Disable this logic block.

        Automatically called when one of the
        disable_event events is posted. Can also manually be called.
        """
        self.debug_log("Disabling")
        self.enabled = False
        self.post_update_event()

    @event_handler(4)
    def event_reset(self, **kwargs):
        """Event handler for reset event."""
        del kwargs
        self.reset()

    def reset(self):
        """Reset the progress towards completion of this logic block.

        Automatically called when one of the reset_event events is called.
        Can also be manually called.
        """
        self.completed = False
        self.value = self.get_start_value()
        self.debug_log("Resetting")
        self.post_update_event()

    @event_handler(5)
    def event_restart(self, **kwargs):
        """Event handler for restart event."""
        del kwargs
        self.restart()

    def restart(self):
        """Restart this logic block by calling reset() and enable().

        Automatically called when one of the restart_event events is called.
        Can also be manually called.
        """
        self.debug_log("Restarting (resetting then enabling)")
        self.reset()
        self.enable()

    def complete(self):
        """Mark this logic block as complete.

        Posts the 'events_when_complete'
        events and optionally restarts this logic block or disables it,
        depending on this block's configuration settings.
        """
        # if already completed do not complete again
        if self.completed:
            return

        # otherwise mark as completed
        self.completed = True

        self.debug_log("Complete")
        if self.config['events_when_complete']:
            for event in self.config['events_when_complete']:
                self.machine.events.post(event)
        '''event: logicblock_(name)_complete

        desc: The logic block called "name" has just been completed.

        Note that this is the default completion event for logic blocks, but
        this can be changed in a logic block's "events_when_complete:" setting,
        so this might not be the actual event that's posted for all logic
        blocks in your machine.
        '''

        # call reset to reset completion
        if self.config['reset_on_complete']:
            self.reset()

        # disable block
        if self.config['disable_on_complete']:
            self.disable()


class Counter(LogicBlock):

    """A type of LogicBlock that tracks multiple hits of a single event.

    This counter can be configured to track hits towards a specific end-goal
    (like number of tilt hits to tilt), or it can be an open-ended count (like
    total number of ramp shots).

    It can also be configured to count up or to count down, and can have a
    configurable counting interval.
    """

    config_section = 'counters'
    collection = 'counters'
    class_label = 'counter'

    __slots__ = ["delay", "ignore_hits", "hit_value"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Initialise counter."""
        super().__init__(machine, name)
        self.delay = DelayManager(self.machine)

        self.ignore_hits = False
        self.hit_value = -1

    async def _initialize(self):
        await super()._initialize()
        self.debug_log("Creating Counter LogicBlock")
        self.hit_value = self.config['count_interval']

        if self.config['direction'] == 'down' and self.hit_value > 0:
            self.hit_value *= -1
        elif self.config['direction'] == 'up' and self.hit_value < 0:
            self.hit_value *= -1
        # Add control events if included in the config
        if self.config['control_events']:
            self._setup_control_events(self.config['control_events'])

    def add_control_events_in_mode(self, mode: Mode) -> None:
        """Do not auto enable this device in modes."""

    def _setup_control_events(self, event_list):
        self.debug_log("Setting up control events")

        kwargs = {}
        for entry in event_list:
            if entry['action'] in ('add', 'subtract', 'jump'):
                handler = getattr(self, "event_{}".format(entry['action']))
                kwargs = {'value': entry['value']}
            else:
                raise AssertionError("Invalid control_event action {} in mode".
                                     format(entry['action']), self.name)
            self.machine.events.add_handler(entry['event'], handler, **kwargs)

    def check_complete(self, count_complete_value=None):
        """Check if counter is completed.

        Return true if the counter has reached or surpassed its specified
        completion value, return False if no completion criteria or is
        not complete.
        """
        # If count_complete_value was not passed, obtain it
        if count_complete_value is None and self.config.get("count_complete_value"):
            count_complete_value = self.config["count_complete_value"].evaluate([])

        if count_complete_value is not None:
            if self.config['direction'] == 'up':
                return self.value >= count_complete_value
            if self.config['direction'] == 'down':
                return self.value <= count_complete_value

        return False

    def event_add(self, value, **kwargs):
        """Add to the value of this counter.

        Args:
        ----
            value: Value to add to the counter.
            kwargs: Additional arguments.
        """
        evaluated_value = value.evaluate_or_none(kwargs)
        if evaluated_value is None:
            self.log.warning("Placeholder %s for counter add did not evaluate with args %s", value, kwargs)
            return
        # Add to the counter the specified value
        self.value += evaluated_value
        # Check if count is complete given the updated value
        if self.check_complete():
            self.complete()

    def event_subtract(self, value, **kwargs):
        """Subtract from the value of this counter.

        Args:
        ----
            value: Value to subtract from the counter.
            kwargs: Additional arguments.
        """
        evaluated_value = value.evaluate_or_none(kwargs)
        if evaluated_value is None:
            self.log.warning("Placeholder %s for counter substract did not evaluate with args %s", value, kwargs)
            return
        # Subtract from the counter the specified value
        self.value -= evaluated_value
        # Check if count is complete given the updated value
        if self.check_complete():
            self.complete()

    def event_jump(self, value, **kwargs):
        """Set the internal value of the counter.

        Args:
        ----
            value: Value to add to jump to.
            kwargs: Additional arguments.
        """
        evaluated_value = value.evaluate_or_none(kwargs)
        if evaluated_value is None:
            self.log.warning("Placeholder %s for counter jump did not evaluate with args %s", value, kwargs)
            return
        # Set the internal value of the counter to the specified value
        self.value = evaluated_value
        # Check if count is complete given the updated value
        if self.check_complete():
            self.complete()

    def get_start_value(self) -> int:
        """Return start count."""
        return self.config['starting_count'].evaluate([])

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None) -> dict:
        """Validate logic block config."""
        if 'events_when_hit' not in config:
            # for compatibility post the same default as previously for
            # counters. This one is deprecated.
            config['events_when_hit'] = ['counter_' + self.name + '_hit']

            # this is the one moving forward
            config['events_when_hit'].append('logicblock_' + self.name + '_hit')

        return super().validate_and_parse_config(config, is_mode_config, debug_prefix)

    @event_handler(0)
    def event_count(self, **kwargs):
        """Event handler for count events."""
        del kwargs
        self.count()

    def count(self):
        """Increase the hit progress towards completion.

        This method is also automatically called when one of the
        ``count_events`` is posted.

        """
        if not self.enabled:
            return

        count_complete_value = self.config['count_complete_value'].evaluate([]) if self.config['count_complete_value']\
            is not None else None

        if not self.ignore_hits:
            self.value += self.hit_value
            self.debug_log("Processing Count change. Total: %s", self.value)

            args = {
                "count": self.value
            }
            if count_complete_value is not None:
                if self.config['direction'] == 'down':
                    args['hits'] = self.get_start_value() - self.value
                    args['remaining'] = self.value - count_complete_value
                else:
                    args['hits'] = self.value - self.get_start_value()
                    args['remaining'] = count_complete_value - self.value

            self._post_hit_events(**args)

            if self.check_complete(count_complete_value):
                self.complete()

            if self.config['multiple_hit_window']:
                self.debug_log("Beginning Ignore Hits")
                self.ignore_hits = True
                self.delay.add(name='ignore_hits_within_window',
                               ms=self.config['multiple_hit_window'],
                               callback=self.stop_ignoring_hits)

    def stop_ignoring_hits(self, **kwargs):
        """Cause the Counter to stop ignoring subsequent hits that occur within the 'multiple_hit_window'.

        Automatically called when the window time expires. Can safely be manually called.
        """
        del kwargs
        self.debug_log("Ending Ignore hits")
        self.ignore_hits = False


class Accrual(LogicBlock):

    """A type of LogicBlock which tracks many different events (steps) towards a goal.

    The steps are able to happen in any order.
    """

    config_section = 'accruals'
    collection = 'accruals'
    class_label = 'accrual'

    __slots__ = []  # type: List[str]

    @property
    def config_section_name(self):
        """Return config section."""
        return "accrual"

    async def _initialize(self):
        await super()._initialize()
        self.debug_log("Creating Accrual LogicBlock")
        self.setup_event_handlers()

    def get_start_value(self) -> List[bool]:
        """Return start states."""
        return [False] * len(self.config['events'])

    def setup_event_handlers(self):
        """Add event handlers."""
        for step, events in enumerate(self.config['events']):
            for event in Util.string_to_event_list(events):
                self.machine.events.add_handler(event, self.hit, step=step)

    @event_handler(0)
    def event_advance_random(self, **kwargs):
        """Event handler for advance random events."""
        del kwargs
        if not self.enabled:
            return

        self.debug_log("Advancing random step in accrual.")
        randomized_values = list(enumerate(self.value))
        shuffle(randomized_values)
        step = None
        for step, state in randomized_values:
            if not state:
                break
        else:
            return

        if step is None:
            return

        # call existing path
        self.hit(step)

    def hit(self, step: int, **kwargs):
        """Increase the hit progress towards completion.

        Automatically called
        when one of the `count_events` is posted. Can also manually be
        called.

        Args:
        ----
            step: Integer of the step number (0 indexed) that was just hit.
        """
        del kwargs
        if not self.enabled:
            return

        self.debug_log("Processing hit for step: %s", step)
        if not self.value[step]:
            self.value[step] = True
            self.debug_log("Status: %s", self.value)
            self._post_hit_events(step=step)

        if self.value.count(True) == len(self.value):
            self.complete()


class Sequence(LogicBlock):

    """A type of LogicBlock which tracks many different events (steps) towards a goal.

    The steps have to happen in order.
    """

    config_section = 'sequences'
    collection = 'sequences'
    class_label = 'sequence'

    __slots__ = []  # type: List[str]

    @property
    def config_section_name(self):
        """Return config section."""
        return "sequence"

    async def _initialize(self):
        """Initialise sequence."""
        await super()._initialize()
        self.debug_log("Creating Sequence LogicBlock")
        self.setup_event_handlers()

    def get_start_value(self) -> int:
        """Return start step."""
        return 0

    def setup_event_handlers(self):
        """Add the handlers for the current step."""
        for step, events in enumerate(self.config['events']):
            for event in Util.string_to_event_list(events):
                # increase priority with steps to prevent advancing multiple steps at once
                self.machine.events.add_handler(event, self.hit, step=step, priority=step)

    def hit(self, step: int = None, **kwargs):
        """Increase the hit progress towards completion.

        Automatically called
        when one of the `count_events` is posted. Can also manually be
        called.
        """
        del kwargs
        if not self.enabled:
            return

        if step is not None and step != self.value:
            # got this for another state
            return

        self.debug_log("Processing Hit")

        self.value += 1
        self._post_hit_events(step=self.value)

        if self.value >= len(self.config['events']):
            self.complete()



"""MPF Hardware Service for VPE.

This is separated from the platform because we need to catch a syntax error in python 3.5 and earlier.
"""
import asyncio
from mpf.platforms.visual_pinball_engine import platform_pb2_grpc
from mpf.platforms.visual_pinball_engine import platform_pb2


class MpfHardwareService(platform_pb2_grpc.MpfHardwareServiceServicer):

    """MPF Service for VPE."""

    __slots__ = ["machine", "platform", "switch_queue", "command_queue", "_started"]

    def __init__(self, machine, platform):
        """Initialise MPF service for VPE."""
        self._connected = asyncio.Future()
        self.machine = machine
        self.platform = platform
        self.switch_queue = asyncio.Queue()
        self.command_queue = asyncio.Queue()
        self._started = asyncio.Future()

    def send_command(self, command):
        """Send command to VPE."""
        self.command_queue.put_nowait(command)

    def get_switch_queue(self):
        """Return switch queue."""
        return self.switch_queue

    async def wait_for_vpe_connect(self):
        """Wait until VPE has connected."""
        return await self._connected

    def set_ready(self):
        """Mark service as ready."""
        self._started.set_result(True)

    async def Start(self, request, context):    # noqa
        """Start MPF."""
        self._connected.set_result(request)
        while True:
            command = await self.command_queue.get()
            # this only works in Python 3.6+
            yield command

    async def GetMachineDescription(self, request, context):    # noqa
        """Get Platform Configuration of VPE platform."""
        switches = []
        await self._started
        for switch in self.platform.get_configured_switches():
            switch_description = platform_pb2.SwitchDescription()
            switch_description.name = switch.config.name
            switch_description.hardware_number = switch.number
            switch_description.switch_type = "NC" if switch.config.invert else "NO"
            switches.append(switch_description)

        coils = []
        for coil in self.platform.get_configured_coils():
            coil_description = platform_pb2.CoilDescription()
            coil_description.name = coil.config.name
            coil_description.hardware_number = coil.number
            coils.append(coil_description)

        lights = []
        for light in self.platform.get_configured_lights():
            light_description = platform_pb2.LightDescription()
            light_description.name = light.config.name
            light_description.hardware_channel_number = light.number
            light_description.hardware_channel_color = light.config.color.name
            lights.append(light_description)

        dmds = []
        for dmd in self.platform.get_configured_dmds():
            dmd_description = platform_pb2.DmdDescription()
            dmd_description.name = dmd.name
            if dmd.color_mapping == "RGB":
                dmd_description.color_mapping = platform_pb2.DmdDescription.ColorMapping.RGB
            elif dmd.color_mapping == "BW":
                dmd_description.color_mapping = platform_pb2.DmdDescription.ColorMapping.BW
            else:
                raise AssertionError("Invalid color mapping {}".format(dmd.color_mapping))

            dmd_description.height = dmd.height
            dmd_description.width = dmd.width
            dmds.append(dmd_description)

        machine_description = platform_pb2.MachineDescription(switches=switches, coils=coils, lights=lights, dmds=dmds)

        return machine_description

    async def SendSwitchChanges(self, request_iterator, context):   # noqa
        """Process a stream of switches."""
        async for element in request_iterator:
            self.switch_queue.put_nowait(element)

        return platform_pb2.EmptyResponse()

    async def Quit(self, request, context):     # noqa
        """Stop MPF."""
        self.machine.stop(reason="VPE exited.")
        return platform_pb2.EmptyResponse()



"""Clock tests."""

import unittest

import asyncio

from mpf.core.clock import ClockBase
from functools import partial

counter = 0


def callback():
    """Global test cb."""
    global counter
    counter += 1


class ClockTestCase(unittest.TestCase):

    def setUp(self):
        global counter
        counter = 0

        try:
            import uvloop
        except ImportError:
            pass
        else:
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.clock = ClockBase(loop=self.loop)
        self.callback_order = []

    def tearDown(self):
        self.loop.close()

    def advance_time_and_run(self, delta=1.0):
        self.loop.run_until_complete(asyncio.sleep(delay=delta))

    def callback1(self, number):
        self.callback_order.append(number)

    def test_schedule_once(self):
        self.clock.schedule_once(callback)
        self.advance_time_and_run(0.001)
        self.assertEqual(counter, 1)

    def test_schedule_once_with_timeout(self):
        self.clock.schedule_once(callback, .001)
        self.advance_time_and_run(0.002)
        self.assertEqual(counter, 1)

    def test_schedule_once_twice(self):
        self.clock.schedule_once(callback)
        self.clock.schedule_once(callback)
        self.advance_time_and_run(0.001)
        self.assertEqual(counter, 2)

    def test_unschedule(self):
        cb1 = self.clock.schedule_once(callback)
        self.clock.schedule_once(callback)
        self.clock.unschedule(cb1)
        self.advance_time_and_run(0.001)
        self.assertEqual(counter, 1)



from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallDevicesHoldCoil(MpfTestCase):
    def get_config_file(self):
        return 'test_hold_coil.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def setUp(self):
        super().setUp()
        self.machine.ball_controller.num_balls_known = 99
        self.machine.playfield.balls = 10
        self.machine.playfield.available_balls = 10

    def test_holdcoil_with_direct_release(self):
        self.machine.coils['hold_coil'].enable = MagicMock()
        self.machine.coils['hold_coil'].disable = MagicMock()
        # after hold switch was posted it should enable the hold_coil
        self.machine.events.post('test_hold_event')
        self.advance_time_and_run(0.1)
        self.machine.coils['hold_coil'].enable.assert_called_once_with()
        assert not self.machine.coils['hold_coil'].disable.called

        # wait some more. coil should stay active
        self.advance_time_and_run(300)
        self.machine.coils['hold_coil'].enable.assert_called_once_with()
        assert not self.machine.coils['hold_coil'].disable.called

        # we trigger entrance switch
        self.assertEqual(0, self.machine.ball_devices['test'].balls)
        self.machine.coils['hold_coil'].enable = MagicMock()
        self.machine.coils['hold_coil'].disable = MagicMock()
        self.machine.switch_controller.process_switch(name='s_entrance', state=1)
        self.machine.switch_controller.process_switch(name='s_entrance', state=0)

        self.advance_time_and_run(300)
        self.assertEqual(0, self.machine.ball_devices['test'].balls)

        # the device should eject the ball right away because nobody claimed it
        self.machine.coils['hold_coil'].disable.assert_called_once_with()
        assert not self.machine.coils['hold_coil'].enable.called

    def test_holdcoil_which_keeps_ball(self):
        # add one ball
        self.assertEqual(0, self.machine.ball_devices['test2'].balls)
        self.machine.coils['hold_coil2'].enable = MagicMock()
        self.machine.coils['hold_coil2'].disable = MagicMock()
        self.machine.events.post('test_hold_event2')
        self.machine.switch_controller.process_switch(name='s_entrance2', state=1)
        self.machine.switch_controller.process_switch(name='s_entrance2', state=0)

        self.advance_time_and_run(300)
        self.machine.coils['hold_coil2'].enable.assert_called_once_with()
        assert not self.machine.coils['hold_coil2'].disable.called
        self.assertEqual(1, self.machine.ball_devices['test2'].balls)

        # add a second ball
        self.machine.coils['hold_coil2'].enable = MagicMock()
        self.machine.coils['hold_coil2'].disable = MagicMock()
        self.machine.events.post('test_hold_event2')
        self.machine.switch_controller.process_switch(name='s_entrance2', state=1)
        self.machine.switch_controller.process_switch(name='s_entrance2', state=0)
        self.advance_time_and_run(300)
        self.machine_run()
        self.machine.coils['hold_coil2'].enable.assert_called_once_with()
        assert not self.machine.coils['hold_coil2'].disable.called
        self.assertEqual(2, self.machine.ball_devices['test2'].balls)

        # eject one ball
        self.machine.coils['hold_coil2'].enable = MagicMock()
        self.machine.coils['hold_coil2'].disable = MagicMock()
        self.machine.ball_devices['test2'].eject()
        self.advance_time_and_run(0.2)
        self.machine.coils['hold_coil2'].disable.assert_called_once_with()
        assert not self.machine.coils['hold_coil2'].enable.called

        # it should reenable the hold coil after 1s because there is a second ball
        self.machine.coils['hold_coil2'].enable = MagicMock()
        self.machine.coils['hold_coil2'].disable = MagicMock()
        self.advance_time_and_run(2)
        assert not self.machine.coils['hold_coil2'].disable.called
        self.machine.coils['hold_coil2'].enable.assert_called_once_with()
        self.assertEqual(1, self.machine.ball_devices['test2'].balls)

    def test_holdcoil_which_keeps_ball_multiple_entries(self):
        # add one ball
        self.machine.ball_devices['test2'].ball_count_handler.counter._last_count = 1
        self.machine.ball_devices['test2'].available_balls = 1
        self.machine.ball_devices['test2'].ball_count_handler._set_ball_count(1)

        # eject one ball
        self.machine.coils['hold_coil2'].enable = MagicMock()
        self.machine.coils['hold_coil2'].disable = MagicMock()
        self.machine.ball_devices['test2'].eject()
        self.advance_time_and_run(0.2)
        self.machine.coils['hold_coil2'].disable.assert_called_once_with()
        assert not self.machine.coils['hold_coil2'].enable.called

        # during the hold add another ball. it should not enable hold now
        self.machine.coils['hold_coil2'].enable = MagicMock()
        self.machine.coils['hold_coil2'].disable = MagicMock()
        self.machine.events.post('test_hold_event2')
        self.machine.switch_controller.process_switch(name='s_entrance2', state=1)
        self.machine.switch_controller.process_switch(name='s_entrance2', state=0)
        self.advance_time_and_run(0.2)
        assert not self.machine.coils['hold_coil2'].disable.called
        assert not self.machine.coils['hold_coil2'].enable.called

        # it should reenable the hold coil after 1s because there is a second ball
        self.machine.coils['hold_coil2'].enable = MagicMock()
        self.machine.coils['hold_coil2'].disable = MagicMock()
        self.advance_time_and_run(2)
        assert not self.machine.coils['hold_coil2'].disable.called
        self.machine.coils['hold_coil2'].enable.assert_called_once_with()
        self.assertEqual(1, self.machine.ball_devices['test2'].balls)

        # eject that ball. coil should stay off
        self.machine.coils['hold_coil2'].enable = MagicMock()
        self.machine.coils['hold_coil2'].disable = MagicMock()
        self.machine.ball_devices['test2'].eject()
        self.advance_time_and_run(300)
        self.machine.coils['hold_coil2'].disable.assert_called_once_with()
        assert not self.machine.coils['hold_coil2'].enable.called
        self.assertEqual(0, self.machine.ball_devices['test2'].balls)

    def test_holdcoil_with_hold_and_entry_switch(self):
        # add one ball
        self.assertEqual(0, self.machine.ball_devices['test3'].balls)
        self.machine.coils['hold_coil3'].enable = MagicMock()
        self.machine.coils['hold_coil3'].disable = MagicMock()
        self.machine.switch_controller.process_switch(name='s_entrance_and_hold3', state=1)
        self.machine.switch_controller.process_switch(name='s_entrance_and_hold3', state=0)

        self.advance_time_and_run(300)
        self.machine.coils['hold_coil3'].enable.assert_called_once_with()
        assert not self.machine.coils['hold_coil3'].disable.called
        self.assertEqual(1, self.machine.ball_devices['test3'].balls)

        # add a second ball
        self.machine.coils['hold_coil3'].enable = MagicMock()
        self.machine.coils['hold_coil3'].disable = MagicMock()
        self.machine.switch_controller.process_switch(name='s_entrance_and_hold3', state=1)
        self.machine.switch_controller.process_switch(name='s_entrance_and_hold3', state=0)
        self.advance_time_and_run(300)
        self.machine_run()
        self.machine.coils['hold_coil3'].enable.assert_called_once_with()
        assert not self.machine.coils['hold_coil3'].disable.called
        self.assertEqual(2, self.machine.ball_devices['test3'].balls)

        # eject one ball
        self.machine.coils['hold_coil3'].enable = MagicMock()
        self.machine.coils['hold_coil3'].disable = MagicMock()
        self.machine.ball_devices['test3'].eject()
        self.advance_time_and_run(0.2)
        self.machine.coils['hold_coil3'].disable.assert_called_once_with()
        assert not self.machine.coils['hold_coil3'].enable.called

        # it should reenable the hold coil after 1s because there is a second ball
        self.machine.coils['hold_coil3'].enable = MagicMock()
        self.machine.coils['hold_coil3'].disable = MagicMock()
        self.advance_time_and_run(2)
        assert not self.machine.coils['hold_coil3'].disable.called
        self.machine.coils['hold_coil3'].enable.assert_called_once_with()
        self.assertEqual(1, self.machine.ball_devices['test3'].balls)

        # eject another ball
        self.machine.coils['hold_coil3'].enable = MagicMock()
        self.machine.coils['hold_coil3'].disable = MagicMock()
        self.machine.ball_devices['test3'].eject()
        self.advance_time_and_run(0.2)
        self.machine.coils['hold_coil3'].disable.assert_called_once_with()
        assert not self.machine.coils['hold_coil3'].enable.called
        self.assertEqual(0, self.machine.ball_devices['test3'].balls)

        # coil should not reenable
        self.machine.coils['hold_coil3'].enable = MagicMock()
        self.machine.coils['hold_coil3'].disable = MagicMock()
        self.advance_time_and_run(30)
        assert not self.machine.coils['hold_coil3'].enable.called

    def test_holdcoil_with_ball_switches(self):
        # add one ball
        self.assertEqual(0, self.machine.ball_devices['test4'].balls)
        self.machine.coils['hold_coil4'].enable = MagicMock()
        self.machine.coils['hold_coil4'].disable = MagicMock()
        self.machine.switch_controller.process_switch(name='s_ball4_1', state=1)

        self.advance_time_and_run(300)
        self.machine.coils['hold_coil4'].enable.assert_called_once_with()
        assert not self.machine.coils['hold_coil4'].disable.called
        self.assertEqual(1, self.machine.ball_devices['test4'].balls)

        # add a second ball
        self.machine.coils['hold_coil4'].enable = MagicMock()
        self.machine.coils['hold_coil4'].disable = MagicMock()
        self.machine.switch_controller.process_switch(name='s_ball4_2', state=1)
        self.advance_time_and_run(300)
        self.machine_run()
        self.machine.coils['hold_coil4'].enable.assert_called_once_with()
        assert not self.machine.coils['hold_coil4'].disable.called
        self.assertEqual(2, self.machine.ball_devices['test4'].balls)

        # eject one ball
        self.machine.coils['hold_coil4'].enable = MagicMock()
        self.machine.coils['hold_coil4'].disable = MagicMock()
        self.machine.ball_devices['test4'].eject()
        self.advance_time_and_run(0.2)
        self.machine.coils['hold_coil4'].disable.assert_called_once_with()
        assert not self.machine.coils['hold_coil4'].enable.called
        self.machine.switch_controller.process_switch(name='s_ball4_1', state=0)
        self.machine.switch_controller.process_switch(name='s_ball4_2', state=0)
        self.machine.switch_controller.process_switch(name='s_ball4_1', state=1)

        # it should reenable the hold coil after 1s because there is a second ball
        self.machine.coils['hold_coil4'].enable = MagicMock()
        self.machine.coils['hold_coil4'].disable = MagicMock()
        self.advance_time_and_run(2)
        assert not self.machine.coils['hold_coil4'].disable.called
        self.machine.coils['hold_coil4'].enable.assert_called_once_with()
        self.assertEqual(1, self.machine.ball_devices['test4'].balls)

    def test_holdcoil_with_ball_switches_eject_fail(self):
        # add one ball
        self.assertEqual(0, self.machine.ball_devices['test4'].balls)
        self.machine.coils['hold_coil4'].enable = MagicMock()
        self.machine.coils['hold_coil4'].disable = MagicMock()
        self.machine.switch_controller.process_switch(name='s_ball4_1', state=1)

        self.advance_time_and_run(300)
        self.machine.coils['hold_coil4'].enable.assert_called_once_with()
        assert not self.machine.coils['hold_coil4'].disable.called
        self.assertEqual(1, self.machine.ball_devices['test4'].balls)

        # eject one ball
        self.machine.coils['hold_coil4'].enable = MagicMock()
        self.machine.coils['hold_coil4'].disable = MagicMock()
        self.machine.ball_devices['test4'].eject()
        self.advance_time_and_run(0.2)
        self.machine.coils['hold_coil4'].disable.assert_called_once_with()
        assert not self.machine.coils['hold_coil4'].enable.called
        self.machine.coils['hold_coil4'].disable = MagicMock()

        # after 2s the coil should get enabled again because there is still a ball in the device
        self.advance_time_and_run(2)
        self.machine.coils['hold_coil4'].enable.assert_called_once_with()
        assert not self.machine.coils['hold_coil4'].disable.called
        self.machine.coils['hold_coil4'].enable = MagicMock()

        # no ball switches change. eject should fail
        self.advance_time_and_run(8)

        # it ejects again
        self.machine.coils['hold_coil4'].disable.assert_called_once_with()
        assert not self.machine.coils['hold_coil4'].enable.called

        # ball leaves
        self.machine.switch_controller.process_switch(name='s_ball4_1', state=0)
        self.advance_time_and_run(1)
        self.assertEqual(0, self.machine.ball_devices['test4'].balls)
        self.advance_time_and_run(10)
        self.assertEqual("idle", self.machine.ball_devices['test4']._state)
        assert not self.machine.coils['hold_coil4'].enable.called




"""Light config player."""
from typing import List

from mpf.config_players.device_config_player import DeviceConfigPlayer
from mpf.core.rgb_color import RGBColor, ColorException
from mpf.core.utility_functions import Util


class LightPlayer(DeviceConfigPlayer):

    """Sets lights based on config."""

    config_file_section = 'light_player'
    show_section = 'lights'
    machine_collection_name = 'lights'
    allow_placeholders_in_keys = True

    __slots__ = []  # type: List[str]

    # pylint: disable-msg=too-many-locals
    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Set light color based on config."""
        key = kwargs.get("key", "")
        instance_dict = self._get_instance_dict(context)
        full_context = self._get_full_context(context + key)
        start_time = kwargs.get("start_time", None)

        for light, s in settings.items():
            final_priority = s["priority"]
            try:
                final_priority += priority
            except KeyError:
                final_priority = priority
            if isinstance(light, str):
                light_names = Util.string_to_event_list(light)
                for light_name in light_names:
                    # skip non-replaced placeholders
                    if not light_name or light_name[0:1] == "(" and light_name[-1:] == ")":
                        continue
                    self._light_named_color(light_name, instance_dict, full_context, s['color'], s["fade"],
                                            final_priority, start_time)
            else:
                self._light_color(light, instance_dict, full_context, s['color'], s["fade"], final_priority, start_time)

    def _remove(self, settings, context, key=""):
        instance_dict = self._get_instance_dict(context)
        full_context = self._get_full_context(context + key)

        for light, s in settings.items():
            if isinstance(light, str):
                light_names = Util.string_to_event_list(light)
                for light_name in light_names:
                    self._light_remove_named(light_name, instance_dict, full_context, s['fade'])
            else:
                self._light_remove(light, instance_dict, full_context, s['fade'])

    def _light_remove_named(self, light_name, instance_dict, full_context, fade_ms):
        try:
            lights = [self.machine.lights[light_name]]
        except KeyError:
            lights = self.machine.lights.items_tagged(light_name)

        for light in lights:
            self._light_remove(light, instance_dict, full_context, fade_ms)

    @staticmethod
    def _light_remove(light, instance_dict, full_context, fade_ms):
        light.remove_from_stack_by_key(full_context, fade_ms)
        try:
            del instance_dict[(full_context, light)]
        except KeyError:
            pass

    # pylint: disable-msg=too-many-arguments
    def handle_subscription_change(self, value, settings, priority, context, key):
        """Handle subscriptions."""
        if value:
            self.play(settings, context, "", priority, key=key)
        else:
            self._remove(settings, context, key=key)

    # pylint: disable-msg=too-many-arguments
    def _light_named_color(self, light_name, instance_dict,
                           full_context, color, fade_ms, priority, start_time):
        try:
            lights = [self.machine.lights[light_name]]
        except KeyError:
            lights = self.machine.lights.items_tagged(light_name)

        if not lights:
            raise AssertionError("Could not find light or tag {} in {}".format(light_name, full_context))

        for light in lights:
            self._light_color(light, instance_dict, full_context, color, fade_ms, priority, start_time)

    # pylint: disable-msg=too-many-arguments
    def _light_color(self, light, instance_dict, full_context, color, fade_ms, priority, start_time):
        if isinstance(color, str) and color == "stop":
            self._light_remove(light, instance_dict, full_context, fade_ms)
            return
        if isinstance(color, str) and color != "on":
            color = self._convert_color(color, context=light)
        light.color(color, key=full_context, fade_ms=fade_ms, priority=priority, start_time=start_time)
        instance_dict[(full_context, light)] = light

    def clear_context(self, context):
        """Remove all colors which were set in context."""
        for (full_context, _), light in self._get_instance_dict(context).items():
            light.remove_from_stack_by_key(full_context)

        self._reset_instance_dict(context)

    def _convert_color(self, color, *, context=None) -> RGBColor:
        """Convert color to RGBColor."""
        # hack to keep compatibility for matrix_light values
        if len(color) == 1:
            color = "0" + color + "0" + color + "0" + color
        elif len(color) == 2:
            color = color + color + color

        try:
            color = RGBColor(color)
        except ColorException as e:
            self.raise_config_error("Invalid color {}".format(color), 1, source_exception=e, context=context)

        return color

    def _expand_device_config(self, device_settings):
        # convert all colors to RGBColor
        device_settings = super()._expand_device_config(device_settings)
        color = device_settings['color']
        if isinstance(color, str) and "(" not in color and color not in ("on", "stop"):
            device_settings['color'] = self._convert_color(color)

        return device_settings

    def get_express_config(self, value):
        """Parse express config."""
        value = str(value).replace(' ', '')
        fade = None
        if '-f' in value:
            # Value contains both a color value and a fade value, parse it into
            # its individual components
            composite_value = value.split('-f')
            value = composite_value[0]
            fade = Util.string_to_ms(composite_value[1])

        return dict(color=value, fade=fade)



"""BCP module."""
import asyncio
from functools import partial
from typing import List

from mpf.core.events import QueuedEvent
from mpf.core.mpf_controller import MpfController

from mpf.core.bcp.bcp_server import BcpServer
from mpf.core.utility_functions import Util

from mpf.core.bcp.bcp_interface import BcpInterface
from mpf.core.bcp.bcp_transport import BcpTransportManager

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class Bcp(MpfController):

    """BCP Module."""

    config_name = "bcp"

    __slots__ = ["interface", "transport", "servers"]

    def __init__(self, machine: "MachineController") -> None:
        """Initialise BCP module."""
        super().__init__(machine)
        self.interface = BcpInterface(machine)
        self.transport = BcpTransportManager(machine)
        self.servers = []       # type: List[BcpServer]

        if self.machine.options['bcp']:
            self.machine.events.add_handler('init_phase_2',
                                            self._setup_bcp_connections)

        self.machine.events.add_handler('init_phase_4',
                                        self._setup_bcp_servers)

        self.machine.events.add_handler('shutdown',
                                        self._stop_servers)

    def send(self, bcp_command, **kwargs):
        """Emulate legacy send.

        Args:
        ----
            bcp_command: Commmand to send
        """
        self.transport.send_to_all_clients(bcp_command, **kwargs)

    def _setup_bcp_connections(self, queue: QueuedEvent, **kwargs):
        """Connect to BCP servers from MPF config."""
        del kwargs
        if ('connections' not in self.machine.config['bcp'] or not
                self.machine.config['bcp']['connections']):
            return

        client_connect_futures = []
        for name, settings in self.machine.config['bcp']['connections'].items():
            self.machine.events.post('bcp_connection_attempt',
                                     name=name,
                                     host=settings['host'],
                                     port=settings['port'])
            '''event: bcp_connection_attempt
            desc: MPF is attempting to make a BCP connection.
            args:
            name: The name of the connection.
            host: The host name MPF is attempting to connect to.
            port: The TCP port MPF is attempting to connect to'''

            client = Util.string_to_class(settings['type'])(self.machine, name, self.machine.bcp)
            client.exit_on_close = settings['exit_on_close']
            connect_future = asyncio.ensure_future(client.connect(settings))
            connect_future.add_done_callback(partial(self.transport.register_transport, client))
            client_connect_futures.append(connect_future)

        # block init until all clients are connected
        if client_connect_futures:
            queue.wait()
            future = asyncio.ensure_future(asyncio.wait(iter(client_connect_futures)))
            future.add_done_callback(lambda x: queue.clear())
            future.add_done_callback(self._bcp_clients_connected)

    def _bcp_clients_connected(self, *args):
        del args
        self.machine.events.post('bcp_clients_connected')
        '''event: bcp_clients_connected
        desc: All BCP outgoing BCP connections have been made.'''

    def _setup_bcp_servers(self, queue: QueuedEvent, **kwargs):
        """Start BCP servers to allow other clients to connect."""
        del kwargs
        if 'servers' not in self.machine.config['bcp'] or not self.machine.config['bcp']['servers']:
            return

        servers_start_futures = []
        for settings in self.machine.config['bcp']['servers'].values():
            server = BcpServer(self.machine, settings['ip'], settings['port'], settings['type'])
            server_future = asyncio.ensure_future(server.start())
            server_future.add_done_callback(lambda x, s=server: self.servers.append(s))     # type: ignore
            servers_start_futures.append(server_future)

        # block init until all servers were started
        if servers_start_futures:
            queue.wait()
            future = asyncio.ensure_future(asyncio.wait(iter(servers_start_futures)))
            future.add_done_callback(lambda x: queue.clear())

    def _stop_servers(self, **kwargs):
        """Stop BCP servers."""
        del kwargs
        for server in self.servers:
            server.stop()



"""Handles all light updates."""
import asyncio

from typing import Dict, Optional

from mpf.core.machine import MachineController
from mpf.core.settings_controller import SettingEntry

from mpf.core.rgb_color import RGBColorCorrectionProfile, RGBColor

from mpf.core.mpf_controller import MpfController
from mpf.core.utility_functions import Util


class LightController(MpfController):

    """Handles light updates and light monitoring."""

    __slots__ = ["light_color_correction_profiles", "_initialised", "_monitor_update_task", "brightness_factor",
                 "_brightness_template"]

    config_name = "light_controller"

    def __init__(self, machine: MachineController) -> None:
        """Initialise lights controller."""
        super().__init__(machine)

        # Generate and add color correction profiles to the machine
        self.light_color_correction_profiles = dict()       # type: Dict[str, RGBColorCorrectionProfile]

        # will only get initialised if there are lights
        self._initialised = False
        self._brightness_template = self.machine.placeholder_manager.build_float_template("machine.brightness", 1.0)
        self._update_brightness()

        self._monitor_update_task = None                    # type: Optional[asyncio.Task]

        if 'named_colors' in self.machine.config:
            self._load_named_colors()

    def _update_brightness(self, *args):
        """Update brightness factor."""
        del args
        self.brightness_factor, future = self._brightness_template.evaluate_and_subscribe([])
        future.add_done_callback(self._update_brightness)

    def _load_named_colors(self):
        """Load named colors from config."""
        for name, color in self.machine.config['named_colors'].items():
            RGBColor.add_color(name, color)

    def initialise_light_subsystem(self):
        """Initialise the light subsystem."""
        if self._initialised:
            return
        self._initialised = True
        self.machine.validate_machine_config_section('light_settings')

        if self.machine.config['light_settings']['color_correction_profiles'] is None:
            self.machine.config['light_settings']['color_correction_profiles'] = (
                dict())

        # Create the default color correction profile and add it to the machine
        default_profile = RGBColorCorrectionProfile.default()
        self.light_color_correction_profiles['default'] = default_profile

        # Add any user-defined profiles specified in the machine config file
        for profile_name, profile_parameters in (
                self.machine.config['light_settings']
                ['color_correction_profiles'].items()):
            self.machine.config_validator.validate_config(
                'color_correction_profile',
                self.machine.config['light_settings']['color_correction_profiles']
                [profile_name], profile_parameters)

            profile = RGBColorCorrectionProfile(profile_name)
            profile.generate_from_parameters(
                gamma=profile_parameters['gamma'],
                whitepoint=profile_parameters['whitepoint'],
                linear_slope=profile_parameters['linear_slope'],
                linear_cutoff=profile_parameters['linear_cutoff'])
            self.light_color_correction_profiles[profile_name] = profile

        # add setting for brightness
        self.machine.settings.add_setting(SettingEntry("brightness", "Brightness", 100, "brightness", 1.0,
                                                       {0.25: "25%", 0.5: "50%", 0.75: "75%", 1.0: "100% (default)"}))

    def monitor_lights(self):
        """Update the color of lights for the monitor."""
        if not self._monitor_update_task:
            self._monitor_update_task = self.machine.clock.loop.create_task(self._monitor_update_lights())
            self._monitor_update_task.add_done_callback(Util.raise_exceptions)

    async def _monitor_update_lights(self):
        colors = {}
        while True:
            for light in self.machine.lights.values():
                color = light.get_color()
                old = colors.get(light, None)
                if old != color:
                    self.machine.device_manager.notify_device_changes(light, "color", old, color)
                    colors[light] = color
            await asyncio.sleep(1 / self.machine.config['mpf']['default_light_hw_update_hz'])



from unittest.mock import MagicMock

from mpf.core.switch_controller import MonitoredSwitchChange

from mpf.tests.MpfTestCase import MpfTestCase


class TestSwitchController(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/switch_controller/'

    def _callback(self, state, ms, switch_name):
        assert(switch_name == "s_test")
        assert(ms == 300)
        assert(state == 1)
        self.isActive = self.machine.switch_controller.is_active(self.machine.switches["s_test"], ms=300)

    def test_monitor(self):
        # add monitor
        monitor = MagicMock()
        self.machine.switch_controller.add_monitor(monitor)

        # hit
        self.hit_switch_and_run("s_test", 1)
        monitor.assert_called_with(MonitoredSwitchChange(
            name='s_test', label='%', platform=self.machine.default_platform, num='1', state=1))
        monitor.reset_mock()

        # release
        self.release_switch_and_run("s_test", 1)
        monitor.assert_called_with(MonitoredSwitchChange(
            name='s_test', label='%', platform=self.machine.default_platform, num='1', state=0))
        monitor.reset_mock()

        # test unknown switch
        self.machine.switch_controller.process_switch_by_num(123123123, 1, self.machine.default_platform)
        monitor.assert_called_with(MonitoredSwitchChange(name='123123123', label='<Platform.Virtual>-123123123',
                                                         platform=self.machine.default_platform, num='123123123',
                                                         state=1))
        monitor.reset_mock()

        # remove monitor
        self.machine.switch_controller.remove_monitor(monitor)

        # no more events
        self.hit_switch_and_run("s_test", 1)
        monitor.assert_not_called()

    def test_wait_futures(self):
        self.hit_switch_and_run("s_test", 1)
        future = self.machine.switch_controller.wait_for_switch(self.machine.switches["s_test"])
        future2 = self.machine.switch_controller.wait_for_switch(self.machine.switches["s_test"], only_on_change=False)
        future3 = self.machine.switch_controller.wait_for_switch(self.machine.switches["s_test"])
        future3.cancel()
        self.assertTrue(future2.done())
        self.release_switch_and_run("s_test", 1)
        self.assertFalse(future.done())
        self.hit_switch_and_run("s_test", 1)
        self.assertTrue(future.done())

    def test_verify_switches(self):
        self.assertTrue(self.machine.clock.loop.run_until_complete(self.machine.switch_controller.verify_switches()))

    def test_is_active_timing(self):
        self.isActive = None

        self.machine.switch_controller.add_switch_handler(
                switch_name="s_test",
                callback=self._callback,
                state=1, ms=300, return_info=True)
        self.machine.switch_controller.process_switch("s_test", 1, True)

        self.advance_time_and_run(3)

        self.assertEqual(True, self.isActive)

    def test_initial_state(self):
        # tests that when MPF starts, the initial states of switches that
        # started in that state are read correctly.
        self.assertFalse(self.machine.switch_controller.is_active(self.machine.switches['s_test'], 1000))

    def _callback_invalid(self):
         raise AssertionError("Should not be called")

    def test_timed_switch_handler(self):
        self.machine.switch_controller.process_switch("s_test", 1)
        self.advance_time_and_run(3)

        self.machine.switch_controller.add_switch_handler(
                switch_name="s_test",
                callback=self._callback_invalid,
                state=0, ms=250)

        self.machine.switch_controller.process_switch("s_test", 0)
        self.advance_time_and_run(.1)

        self.machine.switch_controller.remove_switch_handler(
                switch_name="s_test",
                callback=self._callback_invalid,
                state=0, ms=250)

        self.advance_time_and_run(5)

        cb = MagicMock()
        self.hit_switch_and_run("s_test", .1)
        self.machine.switch_controller.add_switch_handler(
            switch_name="s_test",
            callback=cb,
            state=1, ms=300)
        cb.assert_not_called()
        self.advance_time_and_run(.1)
        cb.assert_not_called()
        self.advance_time_and_run(.2)
        cb.assert_called_with()

        cb = MagicMock()
        self.release_switch_and_run("s_test", .1)
        self.machine.switch_controller.add_switch_handler(
            switch_name="s_test",
            callback=cb,
            state=0, ms=300)
        cb.assert_not_called()
        self.advance_time_and_run(.1)
        cb.assert_not_called()
        self.advance_time_and_run(.2)
        cb.assert_called_with()

    def test_activation_and_deactivation_events(self):
        self.mock_event("test_active")
        self.mock_event("test_active2")
        self.mock_event("test_inactive")
        self.mock_event("test_inactive2")

        self.machine.switch_controller.process_switch("s_test_events", 1)
        self.machine_run()

        self.assertEqual(0, self._events['test_active'])
        self.assertEqual(1, self._events['test_active2'])
        self.assertEqual(0, self._events['test_inactive'])
        self.assertEqual(0, self._events['test_inactive2'])

        self.advance_time_and_run(1)

        self.assertEqual(1, self._events['test_active'])
        self.assertEqual(1, self._events['test_active2'])
        self.assertEqual(0, self._events['test_inactive'])
        self.assertEqual(0, self._events['test_inactive2'])

        self.machine.switch_controller.process_switch("s_test_events", 1)
        self.advance_time_and_run(1)

        self.assertEqual(1, self._events['test_active'])
        self.assertEqual(1, self._events['test_active2'])
        self.assertEqual(0, self._events['test_inactive'])
        self.assertEqual(0, self._events['test_inactive2'])

        self.machine.switch_controller.process_switch("s_test_events", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, self._events['test_active'])
        self.assertEqual(1, self._events['test_active2'])
        self.assertEqual(1, self._events['test_inactive'])
        self.assertEqual(0, self._events['test_inactive2'])

        self.advance_time_and_run(1.1)
        self.assertEqual(1, self._events['test_active'])
        self.assertEqual(1, self._events['test_active2'])
        self.assertEqual(1, self._events['test_inactive'])
        self.assertEqual(1, self._events['test_inactive2'])

    def test_ignore_window_ms(self):
        # first hit. switch becomes active
        self.mock_event("s_test_window_ms_active")
        self.mock_event("s_test_window_ms_inactive")
        self.machine.switch_controller.process_switch("s_test_window_ms", 1)
        self.advance_time_and_run(.001)
        self.assertEventCalled("s_test_window_ms_active")
        self.assertEventNotCalled("s_test_window_ms_inactive")
        self.mock_event("s_test_window_ms_active")

        # and disables
        self.machine.switch_controller.process_switch("s_test_window_ms", 0)
        self.advance_time_and_run(.001)
        self.assertEventNotCalled("s_test_window_ms_active")
        self.assertEventNotCalled("s_test_window_ms_inactive")
        # after the window passed we get an event
        self.advance_time_and_run(.1)
        self.assertEventNotCalled("s_test_window_ms_active")
        self.assertEventCalled("s_test_window_ms_inactive")
        self.mock_event("s_test_window_ms_inactive")

        # and enables again
        self.machine.switch_controller.process_switch("s_test_window_ms", 1)
        self.advance_time_and_run(.01)
        self.assertEventCalled("s_test_window_ms_active")
        self.assertEventNotCalled("s_test_window_ms_inactive")
        self.mock_event("s_test_window_ms_active")

        # disables
        self.machine.switch_controller.process_switch("s_test_window_ms", 0)
        self.advance_time_and_run(.001)
        self.assertEventNotCalled("s_test_window_ms_active")
        self.assertEventNotCalled("s_test_window_ms_inactive")

        # and enables again. nothing happens
        self.machine.switch_controller.process_switch("s_test_window_ms", 1)
        self.advance_time_and_run(.2)
        self.assertEventNotCalled("s_test_window_ms_active")
        self.assertEventNotCalled("s_test_window_ms_inactive")

        # instant response on disable this time
        self.machine.switch_controller.process_switch("s_test_window_ms", 0)
        self.advance_time_and_run(.001)
        self.assertEventNotCalled("s_test_window_ms_active")
        self.assertEventCalled("s_test_window_ms_inactive")

    def test_invert(self):
        self.machine.switch_controller.process_switch("s_test_invert", 1, logical=False)
        self.advance_time_and_run()
        self.assertSwitchState("s_test_invert", 0)

        self.machine.switch_controller.process_switch("s_test_invert", 1, logical=True)
        self.advance_time_and_run()
        self.assertSwitchState("s_test_invert", 1)

        self.machine.switch_controller.process_switch("s_test_invert", 0, logical=False)
        self.advance_time_and_run()
        self.assertSwitchState("s_test_invert", 1)

        self.machine.switch_controller.process_switch("s_test_invert", 0, logical=True)
        self.advance_time_and_run()
        self.assertSwitchState("s_test_invert", 0)

    def _cb1a(self, **kwargs):
        del kwargs
        self.called1 = 1
        self.machine.switch_controller.remove_switch_handler("s_test", self._cb1a)

    def _cb2(self, **kwargs):
        del kwargs
        self.called2 = 1

    def _cb1b(self, **kwargs):
        del kwargs
        self.called1 = 1
        self.machine.switch_controller.remove_switch_handler("s_test", self._cb2)

    def test_remove_in_handler(self):
        self.called1 = 0
        self.called2 = 0
        self.machine.switch_controller.add_switch_handler("s_test", self._cb1a)
        self.machine.switch_controller.add_switch_handler("s_test", self._cb2)

        self.machine.switch_controller.process_switch("s_test", 1)
        self.advance_time_and_run()
        self.assertEqual(1, self.called1)
        self.assertEqual(1, self.called2)

    def test_remove_in_handler2(self):
        self.called1 = 0
        self.called2 = 0
        self.machine.switch_controller.add_switch_handler("s_test", self._cb1b)
        self.machine.switch_controller.add_switch_handler("s_test", self._cb2)

        self.machine.switch_controller.process_switch("s_test", 1)
        self.advance_time_and_run()
        self.assertEqual(1, self.called1)
        self.assertEqual(0, self.called2)

    def _cb3(self, **kwargs):
        del kwargs
        self.called3 = 1

    def test_active_and_inactive_times(self):
        # Regression test for switch_controller bug in _cancel_timed_handlers
        self.machine.switch_controller.add_switch_handler("s_test", self._cb2, ms=5000, state=1)
        self.machine.switch_controller.add_switch_handler("s_test_events", self._cb2, ms=5000, state=1)

        self.called2 = 0
        self.machine.switch_controller.process_switch("s_test", 1)
        self.machine.switch_controller.process_switch("s_test_events", 1)

        self.advance_time_and_run(2)
        self.called2 = 0
        self.machine.switch_controller.process_switch("s_test", 0)
        self.machine.switch_controller.process_switch("s_test_events", 1)

        self.advance_time_and_run(5)
        self.assertEqual(1, self.called2)



"""BCP socket client."""
import json
from urllib.parse import urlsplit, parse_qs, quote, unquote, urlunparse

import asyncio

from typing import Tuple

from mpf._version import __version__, __bcp_version__
from mpf.core.bcp.bcp_client import BaseBcpClient


class MpfJSONEncoder(json.JSONEncoder):

    """Encoder which by default encodes to string."""

    # pylint: disable-msg=method-hidden
    def default(self, o):
        """Encode to string."""
        return str(o)


def decode_command_string(bcp_string) -> Tuple[str, dict]:
    """Decode a BCP command string into separate command and parameter parts.

    Args:
    ----
        bcp_string: The incoming UTF-8, URL encoded BCP command string.

    Returns a tuple of the command string and a dictionary of kwarg pairs.

    Example:
    -------
    Input: trigger?name=hello&foo=Foo%20Bar
    Output: ('trigger', {'name': 'hello', 'foo': 'Foo Bar'})

    Note that BCP commands and parameter names are not case-sensitive and will
    be converted to lowercase. Parameter values are case sensitive, and case
    will be preserved.

    """
    bcp_command = urlsplit(bcp_string, allow_fragments=False)

    if bcp_command.query[0:5] == "json=":
        kwargs = json.loads(bcp_command.query[5:])
        return bcp_command.path, kwargs

    try:
        kwargs = parse_qs(bcp_command.query, keep_blank_values=True)
    except AttributeError:
        kwargs = dict()

    for k, v in kwargs.items():
        if isinstance(v[0], str):
            if v[0].startswith('int:'):
                v[0] = int(v[0][4:])
            elif v[0].startswith('float:'):
                v[0] = float(v[0][6:])
            elif v[0].lower() == 'bool:true':
                v[0] = True
            elif v[0].lower() == 'bool:false':
                v[0] = False
            elif v[0] == 'NoneType:':
                v[0] = None
            else:
                v[0] = unquote(v[0])

            kwargs[k] = v

    return (bcp_command.path,
            dict((k, v[0]) for k, v in kwargs.items()))


def encode_command_string(bcp_command, **kwargs) -> str:
    """Encode a BCP command and kwargs into a valid BCP command string.

    Args:
    ----
        bcp_command: String of the BCP command name.
        **kwargs: Optional pair(s) of kwargs which will be appended to the
            command.

    Returns a string.

    Example:
    -------
    Input: encode_command_string('trigger', {'name': 'hello', 'foo': 'Bar'})
    Output: trigger?name=hello&foo=Bar

    Note that BCP commands and parameter names are not case-sensitive and will
    be converted to lowercase. Parameter values are case sensitive, and case
    will be preserved.

    """
    kwarg_string = ''
    json_needed = False

    for k, v in kwargs.items():
        if isinstance(v, (dict, list)):
            json_needed = True
            break

        value = quote(str(v), '')

        if isinstance(v, bool):  # bool isinstance of int, so this goes first
            value = 'bool:{}'.format(value)
        elif isinstance(v, int):
            value = 'int:{}'.format(value)
        elif isinstance(v, float):
            value = 'float:{}'.format(value)
        elif v is None:
            value = 'NoneType:'
        else:  # cast anything else as a string
            value = str(value)

        kwarg_string += '{}={}&'.format(quote(k, ''),
                                        value)

    kwarg_string = kwarg_string[:-1]

    if json_needed:
        kwarg_string = 'json={}'.format(json.dumps(kwargs, cls=MpfJSONEncoder))

    return str(urlunparse(('', '', bcp_command, '', kwarg_string, '')))


class AsyncioBcpClientSocket():

    """Simple asyncio bcp client."""

    def __init__(self, sender, receiver):
        """Initialise BCP client socket."""
        self._sender = sender
        self._receiver = receiver
        self._receive_buffer = b''

    # pylint: disable-msg=inconsistent-return-statements
    async def read_message(self):
        """Read the next message."""
        while True:
            message = await self._receiver.readline()

            # handle EOF
            if not message:
                raise BrokenPipeError()

            # strip newline
            message = message[0:-1]

            if b'&bytes=' in message:
                message, bytes_needed = message.split(b'&bytes=')
                bytes_needed = int(bytes_needed)

                rawbytes = await self._receiver.readexactly(bytes_needed)

                message_obj = self._process_command(message, rawbytes)

            else:  # no bytes in the message
                message_obj = self._process_command(message)

            if message_obj:
                return message_obj

    def send(self, bcp_command, kwargs):
        """Send a message to the BCP host.

        Args:
        ----
            bcp_command: command to send
            kwargs: parameters to command
        """
        bcp_string = encode_command_string(bcp_command, **kwargs)
        self._sender.write((bcp_string + '\n').encode())

    async def wait_for_response(self, bcp_command):
        """Wait for a command and ignore all others."""
        while True:
            cmd, args = await self.read_message()
            if cmd == "reset":
                self.send("reset_complete", {})
                continue
            if cmd == bcp_command:
                return cmd, args

    @staticmethod
    def _process_command(message, rawbytes=None):
        cmd, kwargs = decode_command_string(message.decode())
        if rawbytes:
            kwargs['rawbytes'] = rawbytes

        return cmd, kwargs


class BCPClientSocket(BaseBcpClient):

    """MPF version of the AsyncioBcpClientSocket.

    (There can be multiple of these to connect to multiple BCP media controllers simultaneously.)

    Args:
    ----
        machine: The main MachineController object.
        name: String name this client.
        bcp: The bcp object.
    """

    config_name = 'bcp_client'

    __slots__ = ["_sender", "_receiver", "_send_goodbye", "_receive_buffer", "_bcp_client_socket_commands", "__dict__"]

    def __init__(self, machine, name, bcp):
        """Initialise BCP client socket."""
        super().__init__(machine, name, bcp)

        self._sender = None
        self._receiver = None
        self._send_goodbye = True
        self._receive_buffer = b''

        self._bcp_client_socket_commands = {'hello': self._receive_hello,
                                            'goodbye': self._receive_goodbye}

    def __repr__(self):
        """Return str representation."""
        return 'BCPClientSocket.{}'.format(self.name)

    async def connect(self, config):
        """Actively connect to server."""
        return await self._setup_client_socket(config['host'], config['port'], config.get('required'))

    async def _setup_client_socket(self, client_host, client_port, required=True):
        """Set up the client socket."""
        self.info_log("Connecting BCP to '%s' at %s:%s...",
                      self.name, client_host, client_port)

        while True:
            connector = self.machine.clock.open_connection(client_host, client_port)
            try:
                self._receiver, self._sender = await connector
            except (ConnectionRefusedError, OSError):
                if required:
                    await asyncio.sleep(.1)
                    continue

                self.info_log("No BCP connection made to '%s' %s:%s",
                              self.name, client_host, client_port)
                return False

            break

        self.info_log("Connected BCP to '%s' %s:%s", self.name, client_host, client_port)

        self.send_hello()
        return True

    def accept_connection(self, receiver, sender):
        """Create client for incoming connection."""
        self._receiver = receiver
        self._sender = sender

        self.send_hello()

    def stop(self):
        """Stop and shut down the socket client."""
        self.debug_log("Stopping socket client")

        if self._send_goodbye:
            self.send_goodbye()

        self._sender.close()

    def send(self, bcp_command, kwargs):
        """Send a message to the BCP host.

        Args:
        ----
            bcp_command: command to send
            kwargs: parameters to command
        """
        try:
            bcp_string = encode_command_string(bcp_command, **kwargs)
        # pylint: disable-msg=broad-except
        except Exception as e:
            self.warning_log("Failed to encode bcp_command %s with args %s. %s", bcp_command, kwargs, e)
            return

        if self._debug:
            self.debug_log('Sending "%s"', bcp_string)

        if hasattr(self._sender.transport, "is_closing") and self._sender.transport.is_closing():
            self.warning_log("Failed to write to bcp since transport is closing. Transport %s", self._sender.transport)
            return
        self._sender.write((bcp_string + '\n').encode())

    # pylint: disable-msg=inconsistent-return-statements
    async def read_message(self):
        """Read the next message."""
        while True:
            message = await self._receiver.readline()

            # handle EOF
            if not message:
                raise BrokenPipeError()

            # strip newline
            message = message[0:-1]

            if b'&bytes=' in message:
                message, bytes_needed = message.split(b'&bytes=')
                bytes_needed = int(bytes_needed)

                rawbytes = await self._receiver.readexactly(bytes_needed)

                message_obj = self._process_command(message, rawbytes)

            else:  # no bytes in the message
                message_obj = self._process_command(message)

            if message_obj:
                return message_obj

    def _process_command(self, message, rawbytes=None):
        if self._debug:
            self.debug_log('Received "%s"', message)

        cmd, kwargs = decode_command_string(message.decode())
        if rawbytes:
            kwargs['rawbytes'] = rawbytes

        if cmd in self._bcp_client_socket_commands:
            self._bcp_client_socket_commands[cmd](**kwargs)
            return None

        return cmd, kwargs

    def _receive_hello(self, **kwargs):
        """Process incoming BCP 'hello' command."""
        self.debug_log('Received BCP Hello from host with kwargs: %s', kwargs)

    def _receive_goodbye(self):
        """Process incoming BCP 'goodbye' command."""
        self._send_goodbye = False
        self.stop()

    def send_hello(self):
        """Send BCP 'hello' command."""
        self.send('hello', {"version": __bcp_version__,
                            "controller_name": 'Mission Pinball Framework',
                            "controller_version": __version__})

    def send_goodbye(self):
        """Send BCP 'goodbye' command."""
        self.send('goodbye', {})



import logging
import asyncio
import grpc
from grpc.experimental import aio

from mpf.platforms.visual_pinball_engine import platform_pb2_grpc
from mpf.platforms.visual_pinball_engine import platform_pb2


class MpfHardwareService(platform_pb2_grpc.MpfHardwareServiceServicer):

    async def Start(self, request, context):
        print("START")
        while True:
            command = platform_pb2.Commands()
            command.fade_light.common_fade_ms = 20
            fade = platform_pb2.FadeLightRequest.ChannelFade()
            fade.light_number = "1"
            fade.target_brightness = 0.5
            command.fade_light.fades.append(fade)
            yield command
            await asyncio.sleep(1)

    async def SendSwitchChanges(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        async for element in request_iterator:
            print(element)

    def Quit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')



async def serve():
    server = aio.server()
    platform_pb2_grpc.add_MpfHardwareServiceServicer_to_server(MpfHardwareService(), server)
    listen_addr = '[::]:50051'
    server.add_insecure_port(listen_addr)
    logging.info("Starting server on %s", listen_addr)
    await server.start()
    await server.wait_for_termination()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    asyncio.run(serve())
